#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the Spark Step Operator."""

import json
import subprocess
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from pydantic import validator
from pyspark.conf import SparkConf

from zenml.config.docker_configuration import DockerConfiguration
from zenml.logger import get_logger
from zenml.repository import Repository
from zenml.step_operators import BaseStepOperator

logger = get_logger(__name__)
if TYPE_CHECKING:
    from zenml.config.resource_configuration import ResourceConfiguration


class SparkStepOperator(BaseStepOperator):
    """Base class for all Spark-related step operators.

    Attributes:
        master: is the master URL for the cluster. You might see different
            schemes for different cluster managers which are supported by Spark
            like Mesos, YARN, or Kubernetes. Within the context of this PR,
            the implementation supports Kubernetes as a cluster manager.
        deploy_mode: can either be 'cluster' (default) or 'client' and it
            decides where the driver node of the application will run.
        submit_kwargs: is the JSON string of a dict, which will be used
            to define additional params if required (Spark has quite a
            lot of different parameters, so including them, all in the step
            operator was not implemented).
    """

    # Instance parameters
    master: str
    deploy_mode: str = "cluster"
    submit_kwargs: Optional[Dict[str, Any]] = None

    @property
    def application_path(self) -> Optional[str]:
        """Optional method for providing the application path.

        This is especially critical when using 'spark-submit' as it defines the
        path (to the application in the environment where Spark is running)
        which is used within the command.

        For more information on how to set this property please check:

        https://spark.apache.org/docs/latest/submitting-applications.html#advanced-dependency-management

        Returns:
            The path to the application entrypoint
        """
        return None

    @validator("submit_kwargs", pre=True)
    def _convert_json_string(
        cls, value: Union[None, str, Dict[str, Any]]
    ) -> Optional[Dict[str, Any]]:
        """Converts potential JSON strings passed via the CLI to dictionaries.

        Args:
            value: The value to convert.

        Returns:
            The converted value.

        Raises:
            TypeError: If the value is not a `str`, `Dict` or `None`.
            ValueError: If the value is an invalid json string or a json string
                that does not decode into a dictionary.
        """
        if isinstance(value, str):
            try:
                dict_ = json.loads(value)
            except json.JSONDecodeError as e:
                raise ValueError(f"Invalid json string '{value}'") from e

            if not isinstance(dict_, Dict):
                raise ValueError(
                    f"Json string '{value}' did not decode into a dictionary."
                )

            return dict_
        elif isinstance(value, Dict) or value is None:
            return value
        else:
            raise TypeError(f"{value} is not a json string or a dictionary.")

    def _resource_configuration(
        self,
        spark_config: SparkConf,
        resource_configuration: "ResourceConfiguration",
    ) -> None:
        """Configures Spark to handle the resource configuration.

        This should serve as the layer between our ResourceConfigurations
        and Spark's own ways of configuring its resources.

        Note: This is still work-in-progress. In the future, we would like to
        enable much more than executor cores and memory with a dedicated
        ResourceConfiguration object.

        Args:
            spark_config: a SparkConf object which collects all the
                configuration parameters
            resource_configuration: the resource configuration for this step
        """
        if resource_configuration.cpu_count:
            spark_config.set(
                "spark.executor.cores",
                str(int(resource_configuration.cpu_count)),
            )

        if resource_configuration.memory:
            # TODO[LOW]: Fix the conversion of the memory unit with a new
            #   type of resource configuration.
            spark_config.set(
                "spark.executor.memory",
                resource_configuration.memory.lower().strip("b"),
            )

    def _backend_configuration(
        self,
        spark_config: SparkConf,
        docker_configuration: "DockerConfiguration",
        pipeline_name: str,
    ) -> None:
        """Configures Spark to handle backends like YARN, Mesos or Kubernetes.

        Args:
            spark_config: a SparkConf object which collects all the
                configuration parameters
            pipeline_name: name of the pipeline which the step to be executed
                is part of
            docker_configuration: the Docker configuration for this step
        """

    def _io_configuration(self, spark_config: SparkConf) -> None:
        """Configures Spark to handle different input/output sources.

        When you work with the Spark integration, you get materializers
        such as SparkDataFrameMaterializer, SparkModelMaterializer. However, in
        many cases, these materializer work only if the environment, where
        Spark is running, is configured according to the artifact store.

        Take s3 as an example. When you want to save a dataframe to an S3
        artifact store, you need to provide configuration parameters such as,
        '"spark.hadoop.fs.s3.impl=org.apache.hadoop.fs.s3a.S3AFileSystem" to
        Spark. This method aims to provide these configuration parameters.

        Args:
            spark_config: a SparkConf object which collects all the
                configuration parameters

        Raises:
            RuntimeError: when the step operator is being used with an S3
                artifact store and the artifact store does not have the
                required authentication
        """
        # Get active artifact store
        repo = Repository()
        artifact_store = repo.active_stack.artifact_store

        from zenml.integrations.s3 import S3_ARTIFACT_STORE_FLAVOR

        # If S3, preconfigure the spark session
        if artifact_store.FLAVOR == S3_ARTIFACT_STORE_FLAVOR:
            (
                key,
                secret,
                _,
            ) = artifact_store._get_credentials()  # type:ignore[attr-defined]
            if key and secret:
                spark_config.setAll(
                    [
                        ("spark.hadoop.fs.s3a.fast.upload", "true"),
                        (
                            "spark.hadoop.fs.s3.impl",
                            "org.apache.hadoop.fs.s3a.S3AFileSystem",
                        ),
                        (
                            "spark.hadoop.fs.AbstractFileSystem.s3.impl",
                            "org.apache.hadoop.fs.s3a.S3A",
                        ),
                        (
                            "spark.hadoop.fs.s3a.aws.credentials.provider",
                            "com.amazonaws.auth.DefaultAWSCredentialsProviderChain",
                        ),
                        ("spark.hadoop.fs.s3a.access.key", f"{key}"),
                        ("spark.hadoop.fs.s3a.secret.key", f"{secret}"),
                    ]
                )
            else:
                raise RuntimeError(
                    "When you use an Spark step operator with an S3 artifact "
                    "store, please make sure that your artifact store has"
                    "defined the required credentials namely the access key "
                    "and the secret access key."
                )
        else:
            logger.warning(
                "In most cases, the Spark step operator requires additional "
                "configuration based on the artifact store flavor you are "
                "using. That also means, that when you use this step operator "
                "with certain artifact store flavor, ZenML can take care of "
                "the pre-configuration. However, the artifact store flavor "
                f"'{artifact_store.FLAVOR}' featured in this stack is not "
                f"known to this step operator and it might require additional "
                f"configuration."
            )

    def _additional_configuration(self, spark_config: SparkConf) -> None:
        """Appends the user-defined configuration parameters.

        Args:
            spark_config: a SparkConf object which collects all the
                configuration parameters
        """
        # Add the additional parameters
        if self.submit_kwargs:
            for k, v in self.submit_kwargs.items():
                spark_config.set(k, v)

    def _launch_spark_job(
        self, spark_config: SparkConf, entrypoint_command: List[str]
    ) -> None:
        """Generates and executes a spark-submit command.

        Args:
            spark_config: a SparkConf object which collects all the
                configuration parameters
            entrypoint_command: a list of strings which is normally used to
                execute the step as a module. However, in the context of this
                step operator, this list is used to form the application
                parameters for spark-submit.

        Raises:
            RuntimeError: if the spark-submit fails
        """
        # Base spark-submit command
        command = [
            f"spark-submit "
            f"--master {self.master} "
            f"--deploy-mode {self.deploy_mode}"
        ]

        # Add the configuration parameters
        command += [f"--conf {c[0]}={c[1]}" for c in spark_config.getAll()]

        # Add the application path
        command.append(self.application_path)  # type: ignore[arg-type]

        # Add the application parameters
        for arg in [
            "--main_module",
            "--step_source_path",
            "--execution_info_path",
            "--input_artifact_types_path",
        ]:
            i = entrypoint_command.index(arg)
            command.extend([arg, entrypoint_command[i + 1]])

        final_command = " ".join(command)

        # Execute the spark-submit
        process = subprocess.Popen(
            final_command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            shell=True,
        )
        stdout, stderr = process.communicate()

        if process.returncode != 0:
            raise RuntimeError(stderr)
        print(stdout)

    def launch(
        self,
        pipeline_name: str,
        run_name: str,
        docker_configuration: "DockerConfiguration",
        entrypoint_command: List[str],
        resource_configuration: "ResourceConfiguration",
    ) -> None:
        """Launches the step on Spark.

        Args:
            pipeline_name: Name of the pipeline which the step to be executed
                is part of.
            run_name: Name of the pipeline run which the step to be executed
                is part of.
            docker_configuration: The Docker configuration for this step.
            entrypoint_command: Command that executes the step.
            resource_configuration: The resource configuration for this step.
        """
        # Start off with an empty configuration
        conf = SparkConf()

        # Add the resource configuration such as cores, memory.
        self._resource_configuration(
            spark_config=conf,
            resource_configuration=resource_configuration,
        )

        # Add the backend configuration such as namespace, docker images names.
        self._backend_configuration(
            spark_config=conf,
            docker_configuration=docker_configuration,
            pipeline_name=pipeline_name,
        )

        # Add the IO configuration for the inputs and the outputs
        self._io_configuration(
            spark_config=conf,
        )

        # Add any additional configuration given by the user.
        self._additional_configuration(
            spark_config=conf,
        )

        # Generate a spark-submit command given the configuration
        self._launch_spark_job(
            spark_config=conf,
            entrypoint_command=entrypoint_command,
        )



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of Docker image builds to run ZenML pipelines."""
import contextlib
import itertools
import os
import subprocess
import sys
from pathlib import PurePath, PurePosixPath
from typing import TYPE_CHECKING, Iterator, List, Optional, Sequence, Tuple

from pydantic import BaseModel

import zenml
from zenml.config.docker_configuration import DockerConfiguration
from zenml.config.global_config import GlobalConfiguration
from zenml.constants import ENV_ZENML_CONFIG_PATH
from zenml.integrations.registry import integration_registry
from zenml.io import fileio
from zenml.logger import get_logger
from zenml.utils import docker_utils, io_utils, source_utils

if TYPE_CHECKING:
    from zenml.runtime_configuration import RuntimeConfiguration
    from zenml.stack import Stack

logger = get_logger(__name__)

DOCKER_IMAGE_WORKDIR = "/app"
DOCKER_IMAGE_ZENML_CONFIG_DIR = ".zenconfig"
DOCKER_IMAGE_ZENML_CONFIG_PATH = (
    f"{DOCKER_IMAGE_WORKDIR}/{DOCKER_IMAGE_ZENML_CONFIG_DIR}"
)

DEFAULT_DOCKER_PARENT_IMAGE = (
    f"zenmldocker/zenml:{zenml.__version__}-"
    f"py{sys.version_info.major}.{sys.version_info.minor}"
)


@contextlib.contextmanager
def _include_active_profile(
    build_context_root: str,
    load_config_path: PurePath = PurePosixPath(DOCKER_IMAGE_ZENML_CONFIG_PATH),
) -> Iterator[None]:
    """Context manager to include the active profile in a Docker build context.

    Args:
        build_context_root: The root of the build context.
        load_config_path: The path of the global configuration inside the
            image.

    Yields:
        None.
    """
    # Save a copy of the current global configuration with the
    # active profile and the active stack configuration into the build
    # context, to have the active profile and active stack accessible from
    # within the container.
    config_path = os.path.join(
        build_context_root, DOCKER_IMAGE_ZENML_CONFIG_DIR
    )
    try:
        GlobalConfiguration().copy_active_configuration(
            config_path, load_config_path=load_config_path
        )
        yield
    finally:
        fileio.rmtree(config_path)


class PipelineDockerImageBuilder(BaseModel):
    """Builds Docker images to run a ZenML pipeline.

    **Usage**:
    ```python
    class MyStackComponent(StackComponent, PipelineDockerImageBuilder):
        def method_that_requires_docker_image(self):
            image_identifier = self.build_and_push_docker_image(...)
            # use the image ID
    ```

    Attributes:
        docker_parent_image: Full name of the Docker image that should be
            used as the parent for the image that will be built. Defaults to
            a ZenML image built for the active Python and ZenML version.

            Additional information to consider:
            * If you specify a custom image here, you need to make sure it has
            ZenML installed.
            * If this is a non-local image, the environment which is running
            the pipeline and building the Docker image needs to be able to pull
            this image.
    """

    docker_parent_image: Optional[str] = None

    @staticmethod
    def _gather_requirements_files(
        docker_configuration: DockerConfiguration, stack: "Stack"
    ) -> List[Tuple[str, str]]:
        """Gathers and/or generates pip requirements files.

        Args:
            docker_configuration: Docker configuration that specifies which
                requirements to install.
            stack: The stack on which the pipeline will run.

        Raises:
            RuntimeError: If the command to export the local python packages
                failed.

        Returns:
            List of tuples (filename, file_content) of all requirements files.
            The files will be in the following order:
            - Packages installed in the local Python environment
            - User-defined requirements
            - Requirements defined by user-defined and/or stack integrations
        """
        requirements_files = []
        logger.info("Gathering requirements for Docker build:")

        # Generate requirements file for the local environment if configured
        if docker_configuration.replicate_local_python_environment:
            try:
                local_requirements = subprocess.check_output(
                    docker_configuration.replicate_local_python_environment.command,
                    shell=True,
                ).decode()
            except subprocess.CalledProcessError as e:
                raise RuntimeError(
                    "Unable to export local python packages."
                ) from e

            requirements_files.append(
                (".zenml_local_requirements", local_requirements)
            )
            logger.info("\t- Including python packages from local environment")

        # Generate/Read requirements file for user-defined requirements
        if isinstance(docker_configuration.requirements, str):
            user_requirements = io_utils.read_file_contents_as_string(
                docker_configuration.requirements
            )
            logger.info(
                "\t- Including user-defined requirements from file `%s`",
                os.path.abspath(docker_configuration.requirements),
            )
        elif isinstance(docker_configuration.requirements, List):
            user_requirements = "\n".join(docker_configuration.requirements)
            logger.info(
                "\t- Including user-defined requirements: %s",
                ", ".join(f"`{r}`" for r in docker_configuration.requirements),
            )
        else:
            user_requirements = None

        if user_requirements:
            requirements_files.append(
                (".zenml_user_requirements", user_requirements)
            )

        # Generate requirements file for all required integrations
        integration_requirements = set(
            itertools.chain.from_iterable(
                integration_registry.select_integration_requirements(
                    integration
                )
                for integration in docker_configuration.required_integrations
            )
        )

        if docker_configuration.install_stack_requirements:
            integration_requirements.update(stack.requirements())

        if integration_requirements:
            integration_requirements_list = sorted(integration_requirements)
            integration_requirements_file = "\n".join(
                integration_requirements_list
            )
            requirements_files.append(
                (
                    ".zenml_integration_requirements",
                    integration_requirements_file,
                )
            )
            logger.info(
                "\t- Including integration requirements: %s",
                ", ".join(f"`{r}`" for r in integration_requirements_list),
            )

        return requirements_files

    @staticmethod
    def _generate_zenml_pipeline_dockerfile(
        parent_image: str,
        docker_configuration: DockerConfiguration,
        requirements_files: Sequence[str] = (),
        entrypoint: Optional[str] = None,
    ) -> List[str]:
        """Generates a Dockerfile.

        Args:
            parent_image: The image to use as parent for the Dockerfile.
            docker_configuration: Docker configuration for this image build.
            requirements_files: Paths of requirements files to install.
            entrypoint: The default entrypoint command that gets executed when
                running a container of an image created by this Dockerfile.

        Returns:
            Lines of the generated Dockerfile.
        """
        lines = [f"FROM {parent_image}", f"WORKDIR {DOCKER_IMAGE_WORKDIR}"]

        if docker_configuration.copy_profile:
            lines.append(
                f"ENV {ENV_ZENML_CONFIG_PATH}={DOCKER_IMAGE_ZENML_CONFIG_PATH}"
            )

        for key, value in docker_configuration.environment.items():
            lines.append(f"ENV {key.upper()}={value}")

        for file in requirements_files:
            lines.append(f"COPY {file} .")
            lines.append(f"RUN pip install --no-cache-dir -r {file}")

        if docker_configuration.copy_files:
            lines.append("COPY . .")
        elif docker_configuration.copy_profile:
            lines.append(f"COPY {DOCKER_IMAGE_ZENML_CONFIG_DIR} .")

        lines.append("RUN chmod -R a+rw .")

        if entrypoint:
            lines.append(f"ENTRYPOINT {entrypoint}")

        return lines

    def build_and_push_docker_image(
        self,
        pipeline_name: str,
        docker_configuration: DockerConfiguration,
        stack: "Stack",
        runtime_configuration: "RuntimeConfiguration",
        entrypoint: Optional[str] = None,
    ) -> str:
        """Builds and pushes a Docker image to run a pipeline.

        Use the image name returned by this method whenever you need to uniquely
        reference the pushed image in order to pull or run it.

        Args:
            pipeline_name: Name of the pipeline for which the image(s) should
                be built.
            docker_configuration: Docker configuration that specifies what
                should be included in the image.
            stack: The stack on which the pipeline will be executed.
            runtime_configuration: The runtime configuration for the pipeline
                run.
            entrypoint: Entrypoint to use for the final image. If left empty,
                no entrypoint will be included in the image.

        Returns:
            The Docker repository digest of the pushed image.

        Raises:
            RuntimeError: If the stack doesn't contain a container registry.
            ValueError: If no Dockerfile and/or custom parent image is
                specified and the Docker configuration doesn't require an
                image build.
        """
        container_registry = stack.container_registry
        if not container_registry:
            raise RuntimeError(
                "Unable to build and push Docker image because stack "
                f"`{stack.name}` has no container registry."
            )

        logger.info(
            "Building Docker image(s) for pipeline `%s`.", pipeline_name
        )
        target_image_name = (
            f"{container_registry.uri}/"
            f"{docker_configuration.target_repository}:{pipeline_name}"
        )

        requires_zenml_build = any(
            [
                docker_configuration.requirements,
                docker_configuration.required_integrations,
                docker_configuration.replicate_local_python_environment,
                docker_configuration.install_stack_requirements,
                docker_configuration.environment,
                docker_configuration.copy_files,
                docker_configuration.copy_profile,
                entrypoint,
            ]
        )

        # Fallback to the value defined on the stack component if the
        # pipeline configuration doesn't have a configured value
        parent_image = (
            docker_configuration.parent_image
            or self.docker_parent_image
            or DEFAULT_DOCKER_PARENT_IMAGE
        )

        if docker_configuration.dockerfile:
            if parent_image != DEFAULT_DOCKER_PARENT_IMAGE:
                logger.warning(
                    "You've specified both a Dockerfile and a custom parent "
                    "image, ignoring the parent image."
                )

            if requires_zenml_build:
                # We will build an additional image on top of this one later
                # to include user files and/or install requirements. The image
                # we build now will be used as the parent for the next build.
                user_image_name = f"zenml-intermediate-build:{pipeline_name}"
                parent_image = user_image_name
            else:
                # The image we'll build from the custom Dockerfile will be
                # used directly, so we tag it with the requested target name.
                user_image_name = target_image_name

            docker_utils.build_image(
                image_name=user_image_name,
                dockerfile=docker_configuration.dockerfile,
                build_context_root=docker_configuration.build_context_root,
                **docker_configuration.build_options,
            )
        elif not requires_zenml_build:
            if parent_image == DEFAULT_DOCKER_PARENT_IMAGE:
                raise ValueError(
                    "Unable to run a ZenML pipeline with the given Docker "
                    "configuration: No Dockerfile or custom parent image "
                    "specified and no files will be copied or requirements "
                    "installed."
                )
            else:
                # The parent image will be used directly to run the pipeline and
                # needs to be tagged so it gets pushed later
                docker_utils.tag_image(parent_image, target=target_image_name)

        if requires_zenml_build:
            requirement_files = self._gather_requirements_files(
                docker_configuration=docker_configuration, stack=stack
            )
            requirements_file_names = [f[0] for f in requirement_files]

            dockerfile = self._generate_zenml_pipeline_dockerfile(
                parent_image=parent_image,
                docker_configuration=docker_configuration,
                requirements_files=requirements_file_names,
                entrypoint=entrypoint,
            )

            if parent_image == DEFAULT_DOCKER_PARENT_IMAGE:
                # The default parent image is static and doesn't require a pull
                # each time
                pull_parent_image = False
            else:
                # If the image is local, we don't need to pull it. Otherwise
                # we play it safe and always pull in case the user pushed a new
                # image for the given name and tag
                pull_parent_image = not docker_utils.is_local_image(
                    parent_image
                )

            # Leave the build context empty if we don't want to copy any files
            requires_build_context = (
                docker_configuration.copy_files
                or docker_configuration.copy_profile
            )
            build_context_root = (
                source_utils.get_source_root_path()
                if requires_build_context
                else None
            )
            maybe_include_active_profile = (
                _include_active_profile(build_context_root=build_context_root)  # type: ignore[arg-type]
                if docker_configuration.copy_profile
                else contextlib.nullcontext()
            )

            with maybe_include_active_profile:
                docker_utils.build_image(
                    image_name=target_image_name,
                    dockerfile=dockerfile,
                    build_context_root=build_context_root,
                    dockerignore=docker_configuration.dockerignore,
                    extra_files=requirement_files,
                    pull=pull_parent_image,
                )

        repo_digest = container_registry.push_image(target_image_name)
        # Store the docker repo digest in the runtime configuration so it gets
        # tracked in the ZenStore
        runtime_configuration["docker_image"] = repo_digest

        return repo_digest



#  Copyright (c) ZenML GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""CLI functionality to interact with pipelines."""


import click

from zenml.cli.cli import TagGroup, cli
from zenml.enums import CliCategories
from zenml.logger import get_logger
from zenml.pipelines.run_pipeline import run_pipeline

logger = get_logger(__name__)


@cli.group(cls=TagGroup, tag=CliCategories.MANAGEMENT_TOOLS)
def pipeline() -> None:
    """Run pipelines."""


@pipeline.command("run", help="Run a pipeline with the given configuration.")
@click.option(
    "--config",
    "-c",
    "config_path",
    type=click.Path(exists=True, dir_okay=False),
    required=True,
)
@click.argument("python_file")
def cli_pipeline_run(python_file: str, config_path: str) -> None:
    """Runs pipeline specified by the given config YAML object.

    Args:
        python_file: Path to the python file that defines the pipeline.
        config_path: Path to configuration YAML file.
    """
    run_pipeline(python_file=python_file, config_path=config_path)



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the Kubernetes Spark Step Operator."""
import os
from typing import Any, ClassVar, Optional

from pyspark.conf import SparkConf

from zenml.config.docker_configuration import DockerConfiguration
from zenml.integrations.spark import SPARK_KUBERNETES_STEP_OPERATOR
from zenml.integrations.spark.step_operators import spark_entrypoint
from zenml.integrations.spark.step_operators.spark_step_operator import (
    SparkStepOperator,
)
from zenml.io.fileio import copy, remove
from zenml.logger import get_logger
from zenml.repository import Repository
from zenml.runtime_configuration import RuntimeConfiguration
from zenml.utils.pipeline_docker_image_builder import (
    DOCKER_IMAGE_WORKDIR,
    PipelineDockerImageBuilder,
)
from zenml.utils.source_utils import get_source_root_path

logger = get_logger(__name__)

LOCAL_ENTRYPOINT = spark_entrypoint.__file__
ENTRYPOINT_NAME = "zenml_spark_entrypoint.py"


class KubernetesSparkStepOperator(
    SparkStepOperator, PipelineDockerImageBuilder
):
    """Step operator which runs Steps with Spark on Kubernetes.

    Attributes:
        namespace: the namespace under which the driver and executor pods
            will run.
        service_account: the service account that will be used by various Spark
            components (to create and watch the pods).
        docker_parent_image: (which originally comes from the
            PipelineDockerImageBuilder base class) indicates the name of a
            base image that has Spark enabled.
    """

    # Parameters for kubernetes
    namespace: Optional[str] = None
    service_account: Optional[str] = None

    # Class configuration
    FLAVOR: ClassVar[str] = SPARK_KUBERNETES_STEP_OPERATOR

    @property
    def application_path(self) -> Any:
        """Provides the application path in the corresponding docker image.

        Returns:
            The path to the application entrypoint within the docker image
        """
        return f"local://{DOCKER_IMAGE_WORKDIR}/{ENTRYPOINT_NAME}"

    def _backend_configuration(
        self,
        spark_config: SparkConf,
        docker_configuration: "DockerConfiguration",
        pipeline_name: str,
    ) -> None:
        """Configures Spark to run on Kubernetes.

        This method will build and push a docker image for the drivers and
        executors and adjust the config accordingly.

        Args:
            spark_config: a SparkConf object which collects all the
                configuration parameters
            docker_configuration: the Docker configuration for this step
            pipeline_name: name of the pipeline which the step to be executed
                is part of

        Raises:
            FileExistsError: if the path where the entrypoint is copied to is
                already occupied
        """
        # Copy the entrypoint
        entrypoint_path = os.path.join(get_source_root_path(), ENTRYPOINT_NAME)

        try:
            copy(LOCAL_ENTRYPOINT, entrypoint_path, overwrite=False)
        except OSError:
            raise FileExistsError(
                f"The Kubernetes Spark step operator needs to copy the step "
                f"entrypoint to {entrypoint_path}, however a file with this "
                f"path already exists."
            )

        try:
            # Build and push the image
            image_name = self.build_and_push_docker_image(
                pipeline_name=pipeline_name,
                docker_configuration=docker_configuration,
                stack=Repository().active_stack,
                runtime_configuration=RuntimeConfiguration(),
            )
        finally:
            remove(entrypoint_path)

        # Adjust the spark configuration
        spark_config.set("spark.kubernetes.container.image", image_name)
        if self.namespace:
            spark_config.set(
                "spark.kubernetes.namespace",
                self.namespace,
            )
        if self.service_account:
            spark_config.set(
                "spark.kubernetes.authenticate.driver.serviceAccountName",
                self.service_account,
            )



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Entrypoint configuration for the Kubernetes worker/step pods."""

from typing import TYPE_CHECKING, Any, List, Set

from zenml.entrypoints import StepEntrypointConfiguration

if TYPE_CHECKING:
    from zenml.steps import BaseStep

RUN_NAME_OPTION = "run_name"


class KubernetesStepEntrypointConfiguration(StepEntrypointConfiguration):
    """Entrypoint configuration for running steps on Kubernetes."""

    @classmethod
    def get_custom_entrypoint_options(cls) -> Set[str]:
        """Kubernetes specific entrypoint options.

        The argument `RUN_NAME_OPTION` is needed for `get_run_name` to have
        consistent values between steps.

        Returns:
            Set of entrypoint options.
        """
        return {RUN_NAME_OPTION}

    @classmethod
    def get_custom_entrypoint_arguments(
        cls, step: "BaseStep", *args: Any, **kwargs: Any
    ) -> List[str]:
        """Kubernetes specific entrypoint arguments.

        Sets the value for the `RUN_NAME_OPTION` argument.

        Args:
            step: ZenML step for which the entrypoint is built.
            args: additional (unused) arguments.
            kwargs: keyword args; needs to include `RUN_NAME_OPTION`.

        Returns:
            List of entrypoint arguments.
        """
        return [
            f"--{RUN_NAME_OPTION}",
            kwargs[RUN_NAME_OPTION],
        ]

    def get_run_name(self, pipeline_name: str) -> str:
        """Returns the ZenML run name.

        Args:
            pipeline_name: Name of the ZenML pipeline (unused).

        Returns:
            ZenML run name.
        """
        job_id: str = self.entrypoint_args[RUN_NAME_OPTION]
        return job_id



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the GCP Secrets Manager."""
import json
import re
from typing import Any, ClassVar, Dict, List, Optional, Tuple

from google.api_core import exceptions as google_exceptions
from google.cloud import secretmanager

from zenml.exceptions import SecretExistsError
from zenml.integrations.gcp import GCP_SECRETS_MANAGER_FLAVOR
from zenml.logger import get_logger
from zenml.secret.base_secret import BaseSecretSchema
from zenml.secret.secret_schema_class_registry import SecretSchemaClassRegistry
from zenml.secrets_managers.base_secrets_manager import (
    ZENML_SECRET_NAME_LABEL,
    BaseSecretsManager,
    SecretsManagerScope,
)
from zenml.secrets_managers.utils import secret_from_dict, secret_to_dict

logger = get_logger(__name__)

ZENML_SCHEMA_NAME = "zenml-schema-name"
ZENML_GROUP_KEY = "zenml-group-key"
ZENML_GCP_SECRET_SCOPE_PATH_SEPARATOR = "-"


def remove_group_name_from_key(combined_key_name: str, group_name: str) -> str:
    """Removes the secret group name from the secret key.

    Args:
        combined_key_name: Full name as it is within the gcp secrets manager
        group_name: Group name (the ZenML Secret name)

    Returns:
        The cleaned key

    Raises:
        RuntimeError: If the group name is not found in the key
    """
    if combined_key_name.startswith(group_name + "_"):
        return combined_key_name[len(group_name + "_") :]
    else:
        raise RuntimeError(
            f"Key-name `{combined_key_name}` does not have the "
            f"prefix `{group_name}`. Key could not be "
            f"extracted."
        )


class GCPSecretsManager(BaseSecretsManager):
    """Class to interact with the GCP secrets manager.

    Attributes:
        project_id:  This is necessary to access the correct GCP project.
                     The project_id of your GCP project space that contains
                     the Secret Manager.
    """

    project_id: str

    # Class configuration
    FLAVOR: ClassVar[str] = GCP_SECRETS_MANAGER_FLAVOR
    SUPPORTS_SCOPING: ClassVar[bool] = True
    CLIENT: ClassVar[Any] = None

    @classmethod
    def _ensure_client_connected(cls) -> None:
        if cls.CLIENT is None:
            cls.CLIENT = secretmanager.SecretManagerServiceClient()

    @classmethod
    def _validate_scope(
        cls,
        scope: SecretsManagerScope,
        namespace: Optional[str],
    ) -> None:
        """Validate the scope and namespace value.

        Args:
            scope: Scope value.
            namespace: Optional namespace value.
        """
        if namespace:
            cls.validate_secret_name_or_namespace(namespace)

    @classmethod
    def validate_secret_name_or_namespace(
        cls,
        name: str,
    ) -> None:
        """Validate a secret name or namespace.

        A Google secret ID is a string with a maximum length of 255 characters
        and can contain uppercase and lowercase letters, numerals, and the
        hyphen (-) and underscore (_) characters. For scoped secrets, we have to
        limit the size of the name and namespace even further to allow space for
        both in the Google secret ID.

        Given that we also save secret names and namespaces as labels, we are
        also limited by the limitation that Google imposes on label values: max
        63 characters and must only contain lowercase letters, numerals
        and the hyphen (-) and underscore (_) characters

        Args:
            name: the secret name or namespace

        Raises:
            ValueError: if the secret name or namespace is invalid
        """
        if not re.fullmatch(r"[a-z0-9_\-]+", name):
            raise ValueError(
                f"Invalid secret name or namespace '{name}'. Must contain "
                f"only lowercase alphanumeric characters and the hyphen (-) and "
                f"underscore (_) characters."
            )

        if name and len(name) > 63:
            raise ValueError(
                f"Invalid secret name or namespace '{name}'. The length is "
                f"limited to maximum 63 characters."
            )

    @property
    def parent_name(self) -> str:
        """Construct the GCP parent path to the secret manager.

        Returns:
            The parent path to the secret manager
        """
        return f"projects/{self.project_id}"

    def _convert_secret_content(
        self, secret: BaseSecretSchema
    ) -> Dict[str, str]:
        """Convert the secret content into a Google compatible representation.

        This method implements two currently supported modes of adapting between
        the naming schemas used for ZenML secrets and Google secrets:

        * for a scoped Secrets Manager, a Google secret is created for each
        ZenML secret with a name that reflects the ZenML secret name and scope
        and a value that contains all its key-value pairs in JSON format.

        * for an unscoped (i.e. legacy) Secrets Manager, this method creates
        multiple Google secret entries for a single ZenML secret by adding the
        secret name to the key name of each secret key-value pair. This allows
        using the same key across multiple secrets. This is only kept for
        backwards compatibility and will be removed some time in the future.

        Args:
            secret: The ZenML secret

        Returns:
            A dictionary with the Google secret name as key and the secret
            contents as value.
        """
        if self.scope == SecretsManagerScope.NONE:
            # legacy per-key secret mapping
            return {f"{secret.name}_{k}": v for k, v in secret.content.items()}

        return {
            self._get_scoped_secret_name(
                secret.name, separator=ZENML_GCP_SECRET_SCOPE_PATH_SEPARATOR
            ): json.dumps(secret_to_dict(secret)),
        }

    def _get_secret_labels(
        self, secret: BaseSecretSchema
    ) -> List[Tuple[str, str]]:
        """Return a list of Google secret label values for a given secret.

        Args:
            secret: the secret object

        Returns:
            A list of Google secret label values
        """
        if self.scope == SecretsManagerScope.NONE:
            # legacy per-key secret labels
            return [
                (ZENML_GROUP_KEY, secret.name),
                (ZENML_SCHEMA_NAME, secret.TYPE),
            ]

        metadata = self._get_secret_metadata(secret)
        return list(metadata.items())

    def _get_secret_scope_filters(
        self,
        secret_name: Optional[str] = None,
    ) -> str:
        """Return a Google filter expression for the entire scope or just a scoped secret.

        These filters can be used when querying the Google Secrets Manager
        for all secrets or for a single secret available in the configured
        scope (see https://cloud.google.com/secret-manager/docs/filtering).

        Args:
            secret_name: Optional secret name to include in the scope metadata.

        Returns:
            Google filter expression uniquely identifying all secrets
            or a named secret within the configured scope.
        """
        if self.scope == SecretsManagerScope.NONE:
            # legacy per-key secret label filters
            if secret_name:
                return f"labels.{ZENML_GROUP_KEY}={secret_name}"
            else:
                return f"labels.{ZENML_GROUP_KEY}:*"

        metadata = self._get_secret_scope_metadata(secret_name)
        filters = [f"labels.{l}={v}" for (l, v) in metadata.items()]
        if secret_name:
            filters.append(f"name:{secret_name}")

        return " AND ".join(filters)

    def _list_secrets(self, secret_name: Optional[str] = None) -> List[str]:
        """List all secrets matching a name.

        This method lists all the secrets in the current scope without loading
        their contents. An optional secret name can be supplied to filter out
        all but a single secret identified by name.

        Args:
            secret_name: Optional secret name to filter for.

        Returns:
            A list of secret names in the current scope and the optional
            secret name.
        """
        self._ensure_client_connected()

        set_of_secrets = set()

        # List all secrets.
        for secret in self.CLIENT.list_secrets(
            request={
                "parent": self.parent_name,
                "filter": self._get_secret_scope_filters(secret_name),
            }
        ):
            if self.scope == SecretsManagerScope.NONE:
                name = secret.labels[ZENML_GROUP_KEY]
            else:
                name = secret.labels[ZENML_SECRET_NAME_LABEL]

            # filter by secret name, if one was given
            if name and (not secret_name or name == secret_name):
                set_of_secrets.add(name)

        return list(set_of_secrets)

    def register_secret(self, secret: BaseSecretSchema) -> None:
        """Registers a new secret.

        Args:
            secret: the secret to register

        Raises:
            SecretExistsError: if the secret already exists
        """
        self.validate_secret_name_or_namespace(secret.name)
        self._ensure_client_connected()

        if self._list_secrets(secret.name):
            raise SecretExistsError(
                f"A Secret with the name {secret.name} already exists"
            )

        adjusted_content = self._convert_secret_content(secret)
        for k, v in adjusted_content.items():
            # Create the secret, this only creates an empty secret with the
            #  supplied name.
            gcp_secret = self.CLIENT.create_secret(
                request={
                    "parent": self.parent_name,
                    "secret_id": k,
                    "secret": {
                        "replication": {"automatic": {}},
                        "labels": self._get_secret_labels(secret),
                    },
                }
            )

            logger.debug("Created empty secret: %s", gcp_secret.name)

            self.CLIENT.add_secret_version(
                request={
                    "parent": gcp_secret.name,
                    "payload": {"data": str(v).encode()},
                }
            )

            logger.debug("Added value to secret.")

    def get_secret(self, secret_name: str) -> BaseSecretSchema:
        """Get a secret by its name.

        Args:
            secret_name: the name of the secret to get

        Returns:
            The secret.

        Raises:
            KeyError: if the secret does not exist
        """
        self.validate_secret_name_or_namespace(secret_name)
        self._ensure_client_connected()

        zenml_secret: Optional[BaseSecretSchema] = None

        if self.scope == SecretsManagerScope.NONE:
            # Legacy secrets are mapped to multiple Google secrets, one for
            # each secret key

            secret_contents = {}
            zenml_schema_name = ""

            # List all secrets.
            for google_secret in self.CLIENT.list_secrets(
                request={
                    "parent": self.parent_name,
                    "filter": self._get_secret_scope_filters(secret_name),
                }
            ):
                secret_version_name = google_secret.name + "/versions/latest"

                response = self.CLIENT.access_secret_version(
                    request={"name": secret_version_name}
                )

                secret_value = response.payload.data.decode("UTF-8")

                secret_key = remove_group_name_from_key(
                    google_secret.name.split("/")[-1], secret_name
                )

                secret_contents[secret_key] = secret_value

                zenml_schema_name = google_secret.labels[ZENML_SCHEMA_NAME]

            if not secret_contents:
                raise KeyError(
                    f"Can't find the specified secret '{secret_name}'"
                )

            secret_contents["name"] = secret_name

            secret_schema = SecretSchemaClassRegistry.get_class(
                secret_schema=zenml_schema_name
            )
            zenml_secret = secret_schema(**secret_contents)

        else:
            # Scoped secrets are mapped 1-to-1 with Google secrets

            google_secret_name = self.CLIENT.secret_path(
                self.project_id,
                self._get_scoped_secret_name(
                    secret_name, separator=ZENML_GCP_SECRET_SCOPE_PATH_SEPARATOR
                ),
            )

            try:
                # fetch the latest secret version
                google_secret = self.CLIENT.get_secret(name=google_secret_name)
            except google_exceptions.NotFound:
                raise KeyError(
                    f"Can't find the specified secret '{secret_name}'"
                )

            # make sure the secret has the correct scope labels to filter out
            # unscoped secrets with similar names
            scope_labels = self._get_secret_scope_metadata(secret_name)
            # all scope labels need to be included in the google secret labels,
            # otherwise the secret does not belong to the current scope
            if not scope_labels.items() <= google_secret.labels.items():
                raise KeyError(
                    f"Can't find the specified secret '{secret_name}'"
                )

            try:
                # fetch the latest secret version
                response = self.CLIENT.access_secret_version(
                    name=f"{google_secret_name}/versions/latest"
                )
            except google_exceptions.NotFound:
                raise KeyError(
                    f"Can't find the specified secret '{secret_name}'"
                )

            secret_value = response.payload.data.decode("UTF-8")
            zenml_secret = secret_from_dict(
                json.loads(secret_value), secret_name=secret_name
            )

        return zenml_secret

    def get_all_secret_keys(self) -> List[str]:
        """Get all secret keys.

        Returns:
            A list of all secret keys
        """
        return self._list_secrets()

    def update_secret(self, secret: BaseSecretSchema) -> None:
        """Update an existing secret by creating new versions of the existing secrets.

        Args:
            secret: the secret to update

        Raises:
            KeyError: if the secret does not exist
        """
        self.validate_secret_name_or_namespace(secret.name)
        self._ensure_client_connected()

        if not self._list_secrets(secret.name):
            raise KeyError(f"Can't find the specified secret '{secret.name}'")

        adjusted_content = self._convert_secret_content(secret)

        for k, v in adjusted_content.items():
            # Create the secret, this only creates an empty secret with the
            #  supplied name.
            google_secret_name = self.CLIENT.secret_path(self.project_id, k)
            payload = {"data": str(v).encode()}

            self.CLIENT.add_secret_version(
                request={"parent": google_secret_name, "payload": payload}
            )

    def delete_secret(self, secret_name: str) -> None:
        """Delete an existing secret by name.

        Args:
            secret_name: the name of the secret to delete

        Raises:
            KeyError: if the secret no longer exists
        """
        self.validate_secret_name_or_namespace(secret_name)
        self._ensure_client_connected()

        if not self._list_secrets(secret_name):
            raise KeyError(f"Can't find the specified secret '{secret_name}'")

        # Go through all gcp secrets and delete the ones with the secret_name
        # as label.
        for secret in self.CLIENT.list_secrets(
            request={
                "parent": self.parent_name,
                "filter": self._get_secret_scope_filters(secret_name),
            }
        ):
            self.CLIENT.delete_secret(request={"name": secret.name})

    def delete_all_secrets(self) -> None:
        """Delete all existing secrets."""
        self._ensure_client_connected()

        # List all secrets.
        for secret in self.CLIENT.list_secrets(
            request={
                "parent": self.parent_name,
                "filter": self._get_secret_scope_filters(),
            }
        ):
            logger.info(f"Deleting Google secret {secret.name}")
            self.CLIENT.delete_secret(request={"name": secret.name})



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the Whylabs step decorator."""

import functools
import os
from typing import Any, Callable, Optional, Type, TypeVar, Union, cast, overload

from zenml.steps import BaseStep
from zenml.steps.utils import STEP_INNER_FUNC_NAME

WHYLABS_LOGGING_ENABLED_ENV = "WHYLABS_LOGGING_ENABLED"
WHYLABS_DATASET_ID_ENV = "WHYLABS_DATASET_ID"

# step entrypoint type
F = TypeVar("F", bound=Callable[..., Any])

# step class type
S = TypeVar("S", bound=Type[BaseStep])


@overload
def enable_whylabs(
    _step: S,
) -> S:
    ...


@overload
def enable_whylabs(
    *,
    dataset_id: Optional[str] = None,
) -> Callable[[S], S]:
    ...


def enable_whylabs(
    _step: Optional[S] = None,
    *,
    dataset_id: Optional[str] = None,
) -> Union[S, Callable[[S], S]]:
    """Decorator to enable Whylabs profiling for a step function.

    Apply this decorator to a ZenML pipeline step to enable Whylabs profiling.

    Note that you also need to have a whylogs Data Validator part of your active
    stack with the Whylabs credentials configured for this to have effect.

    All the whylogs data profile views returned by the step will automatically
    be uploaded to the Whylabs platform if the active whylogs Data Validator
    component is configured with Whylabs credentials, e.g.:

    ```python
    import whylogs as why
    from whylogs.core import DatasetProfileView
    from zenml.integrations.whylogs.whylabs_step_decorator import enable_whylabs

    @enable_whylabs(dataset_id="my_model")
    @step
    def data_loader() -> Output(data=pd.DataFrame, profile=DatasetProfileView,):
        ...
        data = pd.DataFrame(...)
        results = why.log(pandas=dataset)
        profile = results.profile()
        ...
        return data, profile.view()
    ```

    You can also use this decorator with our class-based API like so:
    ```
    @enable_whylabs(dataset_id="my_model")
    class DataLoader(BaseStep):
        def entrypoint(self) -> Output(data=pd.DataFrame, profile=DatasetProfileView,):
            ...
    ```

    Args:
        _step: The decorated step class.
        dataset_id: Optional dataset ID to use for the uploaded profile(s)

    Returns:
        the inner decorator which enhances the input step class with Whylabs
        profiling functionality
    """

    def inner_decorator(_step: S) -> S:
        source_fn = getattr(_step, STEP_INNER_FUNC_NAME)
        new_entrypoint = whylabs_entrypoint(dataset_id)(source_fn)
        if _step._created_by_functional_api():
            # If the step was created by the functional API, the old entrypoint
            # was a static method -> make sure the new one is as well
            new_entrypoint = staticmethod(new_entrypoint)

        setattr(_step, STEP_INNER_FUNC_NAME, new_entrypoint)
        return _step

    if _step is None:
        return inner_decorator
    else:
        return inner_decorator(_step)


def whylabs_entrypoint(
    dataset_id: Optional[str] = None,
) -> Callable[[F], F]:
    """Decorator for a step entrypoint to enable Whylabs logging.

    Apply this decorator to a ZenML pipeline step to enable Whylabs profiling.

    Note that you also need to have a whylogs Data Validator part of your active
    stack with the Whylabs credentials configured for this to have effect.

    All the whylogs data profile views returned by the step will automatically
    be uploaded to the Whylabs platform if the active whylogs Data Validator
    component is configured with Whylabs credentials, e.g.:

    .. highlight:: python
    .. code-block:: python

        import whylogs as why
        from whylogs.core import DatasetProfileView
        from zenml.integrations.whylogs.whylabs_step_decorator import whylabs_entrypoint

        @step
        @whylabs_entrypoint(dataset_id="my_model")
        def data_loader() -> Output(data=pd.DataFrame, profile=DatasetProfileView,):
            ...
            data = pd.DataFrame(...)
            results = why.log(pandas=dataset)
            profile = results.profile()
            ...
            return data, profile.view()

    Args:
        dataset_id: Optional dataset ID to use for the uploaded profile(s)

    Returns:
        the input function enhanced with Whylabs profiling functionality
    """

    def inner_decorator(func: F) -> F:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:  # noqa
            os.environ[WHYLABS_LOGGING_ENABLED_ENV] = "true"
            if dataset_id:
                os.environ[WHYLABS_DATASET_ID_ENV] = dataset_id
            try:
                return func(*args, **kwargs)
            finally:
                del os.environ[WHYLABS_LOGGING_ENABLED_ENV]
                if dataset_id:
                    del os.environ[WHYLABS_DATASET_ID_ENV]

        return cast(F, wrapper)

    return inner_decorator



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of Kubernetes metadata store."""

import os
import subprocess
import sys
import time
from typing import Any, ClassVar, Dict, Union

from kubernetes import client as k8s_client
from ml_metadata.proto import metadata_store_pb2
from ml_metadata.proto.metadata_store_pb2 import MySQLDatabaseConfig
from pydantic import root_validator

from zenml.exceptions import ProvisioningError, StackComponentInterfaceError
from zenml.integrations.kubernetes import KUBERNETES_METADATA_STORE_FLAVOR
from zenml.integrations.kubernetes.orchestrators import kube_utils
from zenml.io import fileio
from zenml.logger import get_logger
from zenml.metadata_stores.base_metadata_store import BaseMetadataStore
from zenml.utils import io_utils, networking_utils

logger = get_logger(__name__)

DEFAULT_KUBERNETES_METADATA_DAEMON_TIMEOUT = 60
DEFAULT_KUBERNETES_METADATA_DAEMON_PID_FILE = "kubernetes_metadata_daemon.pid"
DEFAULT_KUBERNETES_METADATA_DAEMON_LOG_FILE = "kubernetes_metadata_daemon.log"

DEFAULT_KUBERNETES_MYSQL_HOST = "mysql"
DEFAULT_KUBERNETES_MYSQL_LOCAL_HOST = "127.0.0.1"
DEFAULT_KUBERNETES_MYSQL_PORT = 3306
DEFAULT_KUBERNETES_MYSQL_DATABASE = "mysql"
DEFAULT_KUBERNETES_MYSQL_USERNAME = "root"
DEFAULT_KUBERNETES_MYSQL_PASSWORD = ""
DEFAULT_KUBERNETES_MYSQL_SECRET = None


class KubernetesMetadataStore(BaseMetadataStore):
    """Kubernetes metadata store (MySQL database deployed in the cluster).

    Attributes:
        deployment_name: Name of the Kubernetes deployment and corresponding
            service/pod that will be created when calling `provision()`.
        kubernetes_context: Name of the Kubernetes context in which to deploy
            and provision the MySQL database.
        kubernetes_namespace: Name of the Kubernetes namespace.
            Defaults to "default".
        storage_capacity: Storage capacity of the metadata store.
            Defaults to `"10Gi"` (=10GB).
    """

    deployment_name: str
    kubernetes_context: str
    kubernetes_namespace: str = "zenml"
    storage_capacity: str = "10Gi"
    _k8s_core_api: k8s_client.CoreV1Api = None
    _k8s_apps_api: k8s_client.AppsV1Api = None

    FLAVOR: ClassVar[str] = KUBERNETES_METADATA_STORE_FLAVOR

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initiate the Pydantic object and initialize the Kubernetes clients.

        Args:
            *args: The positional arguments to pass to the Pydantic object.
            **kwargs: The keyword arguments to pass to the Pydantic object.
        """
        super().__init__(*args, **kwargs)
        self._initialize_k8s_clients()

    def _initialize_k8s_clients(self) -> None:
        """Initialize the Kubernetes clients."""
        kube_utils.load_kube_config(context=self.kubernetes_context)
        self._k8s_core_api = k8s_client.CoreV1Api()
        self._k8s_apps_api = k8s_client.AppsV1Api()

    @root_validator(skip_on_failure=False)
    def check_required_attributes(
        cls, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Pydantic root_validator.

        This ensures that both `deployment_name` and `kubernetes_context` are
        set and raises an error with a custom error message otherwise.

        Args:
            values: Values passed to the Pydantic constructor.

        Raises:
            StackComponentInterfaceError: if either `deployment_name` or
                `kubernetes_context` is not defined.

        Returns:
            Values passed to the Pydantic constructor.
        """
        usage_note = (
            "Note: the `kubernetes` metadata store flavor is a special "
            "subtype of the `mysql` metadata store that deploys a fresh "
            "MySQL database within your Kubernetes cluster when running "
            "`zenml stack up`. "
            "If you already have a MySQL database running in your cluster "
            "(or elsewhere), simply use the `mysql` metadata store flavor "
            "instead."
        )

        for required_field in ("deployment_name", "kubernetes_context"):
            if required_field not in values:
                raise StackComponentInterfaceError(
                    f"Required field `{required_field}` missing for "
                    "`KubernetesMetadataStore`. " + usage_note
                )

        return values

    @property
    def deployment_exists(self) -> bool:
        """Check whether a MySQL deployment exists in the cluster.

        Returns:
            Whether a MySQL deployment exists in the cluster.
        """
        resp = self._k8s_apps_api.list_namespaced_deployment(
            namespace=self.kubernetes_namespace
        )
        for i in resp.items:
            if i.metadata.name == self.deployment_name:
                return True
        return False

    @property
    def is_provisioned(self) -> bool:
        """If the component provisioned resources to run.

        Checks whether the required MySQL deployment exists.

        Returns:
            True if the component provisioned resources to run.
        """
        return super().is_provisioned and self.deployment_exists

    @property
    def is_running(self) -> bool:
        """If the component is running.

        Returns:
            True if `is_provisioned` else False.
        """
        if sys.platform != "win32":
            from zenml.utils.daemon import check_if_daemon_is_running

            if not check_if_daemon_is_running(self._pid_file_path):
                return False
        else:
            # Daemon functionality is not supported on Windows, so the PID
            # file won't exist. This if clause exists just for mypy to not
            # complain about missing functions
            pass

        return self.is_provisioned

    def provision(self) -> None:
        """Provision the metadata store.

        Creates a deployment with a MySQL database running in it.
        """
        logger.info("Provisioning Kubernetes MySQL metadata store...")
        kube_utils.create_namespace(
            core_api=self._k8s_core_api, namespace=self.kubernetes_namespace
        )
        kube_utils.create_mysql_deployment(
            core_api=self._k8s_core_api,
            apps_api=self._k8s_apps_api,
            namespace=self.kubernetes_namespace,
            storage_capacity=self.storage_capacity,
            deployment_name=self.deployment_name,
        )

        # wait a bit, then make sure deployment pod is alive and running.
        logger.info("Trying to reach Kubernetes MySQL metadata store pod...")
        time.sleep(10)
        kube_utils.wait_pod(
            core_api=self._k8s_core_api,
            pod_name=self.pod_name,
            namespace=self.kubernetes_namespace,
            exit_condition_lambda=kube_utils.pod_is_not_pending,
        )
        logger.info("Kubernetes MySQL metadata store pod is up and running.")

    def deprovision(self) -> None:
        """Deprovision the metadata store by deleting the MySQL deployment."""
        logger.info("Deleting Kubernetes MySQL metadata store...")
        self.suspend()
        kube_utils.delete_deployment(
            apps_api=self._k8s_apps_api,
            deployment_name=self.deployment_name,
            namespace=self.kubernetes_namespace,
        )

    # TODO: code duplication with kubeflow metadata store below.

    @property
    def root_directory(self) -> str:
        """Returns path to the root directory for all files concerning this orchestrator.

        Returns:
            Path to the root directory.
        """
        return os.path.join(
            io_utils.get_global_config_directory(),
            self.FLAVOR,
            str(self.uuid),
        )

    @property
    def _pid_file_path(self) -> str:
        """Returns path to the daemon PID file.

        Returns:
            Path to the daemon PID file.
        """
        return os.path.join(
            self.root_directory, DEFAULT_KUBERNETES_METADATA_DAEMON_PID_FILE
        )

    @property
    def _log_file(self) -> str:
        """Path of the daemon log file.

        Returns:
            Path to the daemon log file.
        """
        return os.path.join(
            self.root_directory, DEFAULT_KUBERNETES_METADATA_DAEMON_LOG_FILE
        )

    def resume(self) -> None:
        """Resumes the metadata store."""
        self.start_metadata_daemon()
        self.wait_until_metadata_store_ready(
            timeout=DEFAULT_KUBERNETES_METADATA_DAEMON_TIMEOUT
        )

    def suspend(self) -> None:
        """Suspends the metadata store."""
        self.stop_metadata_daemon()

    @property
    def pod_name(self) -> str:
        """Name of the Kubernetes pod where the MySQL database is deployed.

        Returns:
            Name of the Kubernetes pod.
        """
        pod_list = self._k8s_core_api.list_namespaced_pod(
            namespace=self.kubernetes_namespace,
            label_selector=f"app={self.deployment_name}",
        )
        return pod_list.items[0].metadata.name  # type: ignore[no-any-return]

    @property
    def host(self) -> str:
        """Get the MySQL host required to access the metadata store.

        This overwrites the MySQL host to use local host when
        running outside of the cluster so we can access the metadata store
        locally for post execution.

        Raises:
            RuntimeError: If the metadata store is not running.

        Returns:
            MySQL host.
        """
        if kube_utils.is_inside_kubernetes():
            return DEFAULT_KUBERNETES_MYSQL_HOST
        if not self.is_running:
            raise RuntimeError(
                "The Kubernetes metadata daemon is not running. Please run the "
                "following command to start it first:\n\n"
                "    'zenml metadata-store up'\n"
            )
        return DEFAULT_KUBERNETES_MYSQL_LOCAL_HOST

    @property
    def port(self) -> int:
        """Get the MySQL port required to access the metadata store.

        Returns:
            int: MySQL port.
        """
        return DEFAULT_KUBERNETES_MYSQL_PORT

    def get_tfx_metadata_config(
        self,
    ) -> Union[
        metadata_store_pb2.ConnectionConfig,
        metadata_store_pb2.MetadataStoreClientConfig,
    ]:
        """Return tfx metadata config for the Kubernetes metadata store.

        Returns:
            The tfx metadata config.
        """
        config = MySQLDatabaseConfig(
            host=self.host,
            port=self.port,
            database=DEFAULT_KUBERNETES_MYSQL_DATABASE,
            user=DEFAULT_KUBERNETES_MYSQL_USERNAME,
            password=DEFAULT_KUBERNETES_MYSQL_PASSWORD,
        )
        connection_config = metadata_store_pb2.ConnectionConfig(mysql=config)
        return connection_config

    def start_metadata_daemon(self) -> None:
        """Starts a daemon process that forwards ports.

        This is so the MySQL database in the Kubernetes cluster is accessible
        on the localhost.

        Raises:
            ProvisioningError: if the daemon fails to start.
        """
        command = [
            "kubectl",
            "--context",
            self.kubernetes_context,
            "--namespace",
            self.kubernetes_namespace,
            "port-forward",
            f"svc/{self.deployment_name}",
            f"{self.port}:{self.port}",
        ]
        if sys.platform == "win32":
            logger.warning(
                "Daemon functionality not supported on Windows. "
                "In order to access the Kubernetes Metadata locally, "
                "please run '%s' in a separate command line shell.",
                self.port,
                " ".join(command),
            )
        elif not networking_utils.port_available(self.port):
            raise ProvisioningError(
                f"Unable to port-forward Kubernetes Metadata to local "
                f"port {self.port} because the port is occupied. In order to "
                f"access the Kubernetes Metadata locally, please "
                f"change the metadata store configuration to use an available "
                f"port or stop the other process currently using the port."
            )
        else:
            from zenml.utils import daemon

            def _daemon_function() -> None:
                """Forwards the port of the Kubernetes metadata store pod ."""
                subprocess.check_call(command)

            daemon.run_as_daemon(
                _daemon_function,
                pid_file=self._pid_file_path,
                log_file=self._log_file,
            )
            logger.info(
                "Started Kubernetes Metadata daemon (check the daemon"
                "logs at %s in case you're not able to access the pipeline"
                "metadata).",
                self._log_file,
            )

    def stop_metadata_daemon(self) -> None:
        """Stops the Kubernetes metadata daemon process if it is running."""
        if sys.platform != "win32" and fileio.exists(self._pid_file_path):
            from zenml.utils import daemon

            daemon.stop_daemon(self._pid_file_path)
            fileio.remove(self._pid_file_path)

    def wait_until_metadata_store_ready(self, timeout: int) -> None:
        """Waits until the metadata store connection is ready.

        Potentially an irrecoverable error could occur or the timeout could
        expire, so it checks for this.

        Args:
            timeout: The maximum time to wait for the metadata store to be
                ready.

        Raises:
            RuntimeError: if the metadata store is not ready after the timeout
        """
        logger.info(
            "Waiting for the Kubernetes metadata store to be ready (this "
            "might take a few minutes)."
        )
        while True:
            try:
                # it doesn't matter what we call here as long as it exercises
                # the MLMD connection
                self.get_pipelines()
                break
            except Exception as e:
                logger.info(
                    "The Kubernetes metadata store is not ready yet. Waiting "
                    "for 10 seconds..."
                )
                if timeout <= 0:
                    raise RuntimeError(
                        f"An unexpected error was encountered while waiting "
                        f"for the Kubernetes metadata store to be functional: "
                        f"{str(e)}"
                    ) from e
                timeout -= 10
                time.sleep(10)

        logger.info("The Kubernetes metadata store is functional.")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['zenml',
 'zenml.alerter',
 'zenml.annotators',
 'zenml.artifact_stores',
 'zenml.artifacts',
 'zenml.cli',
 'zenml.config',
 'zenml.container_registries',
 'zenml.data_validators',
 'zenml.entrypoints',
 'zenml.experiment_trackers',
 'zenml.feature_stores',
 'zenml.integrations',
 'zenml.integrations.airflow',
 'zenml.integrations.airflow.orchestrators',
 'zenml.integrations.aws',
 'zenml.integrations.aws.container_registries',
 'zenml.integrations.aws.secrets_managers',
 'zenml.integrations.aws.step_operators',
 'zenml.integrations.azure',
 'zenml.integrations.azure.artifact_stores',
 'zenml.integrations.azure.secrets_managers',
 'zenml.integrations.azure.step_operators',
 'zenml.integrations.dash',
 'zenml.integrations.dash.visualizers',
 'zenml.integrations.deepchecks',
 'zenml.integrations.deepchecks.data_validators',
 'zenml.integrations.deepchecks.materializers',
 'zenml.integrations.deepchecks.steps',
 'zenml.integrations.deepchecks.visualizers',
 'zenml.integrations.evidently',
 'zenml.integrations.evidently.data_validators',
 'zenml.integrations.evidently.materializers',
 'zenml.integrations.evidently.steps',
 'zenml.integrations.evidently.visualizers',
 'zenml.integrations.facets',
 'zenml.integrations.facets.visualizers',
 'zenml.integrations.feast',
 'zenml.integrations.feast.feature_stores',
 'zenml.integrations.gcp',
 'zenml.integrations.gcp.artifact_stores',
 'zenml.integrations.gcp.orchestrators',
 'zenml.integrations.gcp.secrets_manager',
 'zenml.integrations.gcp.step_operators',
 'zenml.integrations.github',
 'zenml.integrations.github.orchestrators',
 'zenml.integrations.github.secrets_managers',
 'zenml.integrations.graphviz',
 'zenml.integrations.graphviz.visualizers',
 'zenml.integrations.great_expectations',
 'zenml.integrations.great_expectations.data_validators',
 'zenml.integrations.great_expectations.materializers',
 'zenml.integrations.great_expectations.steps',
 'zenml.integrations.great_expectations.visualizers',
 'zenml.integrations.huggingface',
 'zenml.integrations.huggingface.materializers',
 'zenml.integrations.kserve',
 'zenml.integrations.kserve.custom_deployer',
 'zenml.integrations.kserve.model_deployers',
 'zenml.integrations.kserve.secret_schemas',
 'zenml.integrations.kserve.services',
 'zenml.integrations.kserve.steps',
 'zenml.integrations.kubeflow',
 'zenml.integrations.kubeflow.metadata_stores',
 'zenml.integrations.kubeflow.orchestrators',
 'zenml.integrations.kubernetes',
 'zenml.integrations.kubernetes.metadata_stores',
 'zenml.integrations.kubernetes.orchestrators',
 'zenml.integrations.label_studio',
 'zenml.integrations.label_studio.annotators',
 'zenml.integrations.label_studio.label_config_generators',
 'zenml.integrations.label_studio.steps',
 'zenml.integrations.lightgbm',
 'zenml.integrations.lightgbm.materializers',
 'zenml.integrations.mlflow',
 'zenml.integrations.mlflow.experiment_trackers',
 'zenml.integrations.mlflow.model_deployers',
 'zenml.integrations.mlflow.services',
 'zenml.integrations.mlflow.steps',
 'zenml.integrations.neural_prophet',
 'zenml.integrations.neural_prophet.materializers',
 'zenml.integrations.pillow',
 'zenml.integrations.pillow.materializers',
 'zenml.integrations.plotly',
 'zenml.integrations.plotly.visualizers',
 'zenml.integrations.pytorch',
 'zenml.integrations.pytorch.materializers',
 'zenml.integrations.pytorch_lightning',
 'zenml.integrations.pytorch_lightning.materializers',
 'zenml.integrations.s3',
 'zenml.integrations.s3.artifact_stores',
 'zenml.integrations.scipy',
 'zenml.integrations.scipy.materializers',
 'zenml.integrations.seldon',
 'zenml.integrations.seldon.custom_deployer',
 'zenml.integrations.seldon.model_deployers',
 'zenml.integrations.seldon.secret_schemas',
 'zenml.integrations.seldon.services',
 'zenml.integrations.seldon.steps',
 'zenml.integrations.sklearn',
 'zenml.integrations.sklearn.materializers',
 'zenml.integrations.slack',
 'zenml.integrations.slack.alerters',
 'zenml.integrations.slack.steps',
 'zenml.integrations.spark',
 'zenml.integrations.spark.materializers',
 'zenml.integrations.spark.step_operators',
 'zenml.integrations.tekton',
 'zenml.integrations.tekton.orchestrators',
 'zenml.integrations.tensorflow',
 'zenml.integrations.tensorflow.materializers',
 'zenml.integrations.tensorflow.services',
 'zenml.integrations.tensorflow.visualizers',
 'zenml.integrations.vault',
 'zenml.integrations.vault.secrets_manager',
 'zenml.integrations.wandb',
 'zenml.integrations.wandb.experiment_trackers',
 'zenml.integrations.whylogs',
 'zenml.integrations.whylogs.data_validators',
 'zenml.integrations.whylogs.materializers',
 'zenml.integrations.whylogs.secret_schemas',
 'zenml.integrations.whylogs.steps',
 'zenml.integrations.whylogs.visualizers',
 'zenml.integrations.xgboost',
 'zenml.integrations.xgboost.materializers',
 'zenml.io',
 'zenml.materializers',
 'zenml.metadata_stores',
 'zenml.model_deployers',
 'zenml.orchestrators',
 'zenml.orchestrators.local',
 'zenml.pipelines',
 'zenml.post_execution',
 'zenml.secret',
 'zenml.secret.schemas',
 'zenml.secrets_managers',
 'zenml.secrets_managers.local',
 'zenml.services',
 'zenml.services.local',
 'zenml.stack',
 'zenml.step_operators',
 'zenml.steps',
 'zenml.steps.step_interfaces',
 'zenml.utils',
 'zenml.visualizers',
 'zenml.zen_server',
 'zenml.zen_stores',
 'zenml.zen_stores.models']

package_data = \
{'': ['*']}

install_requires = \
['analytics-python>=1.4.0,<2.0.0',
 'apache-beam>=2.30.0,<3.0.0',
 'click-params>=0.3.0,<0.4.0',
 'click>=8.0.1,<9.0.0',
 'distro>=1.6.0,<2.0.0',
 'gitpython>=3.1.18,<4.0.0',
 'httplib2>=0.19.1,<0.20',
 'markupsafe==1.1.1',
 'ml-pipelines-sdk==1.8.0',
 'nbconvert==6.4.4',
 'pandas>=1.1.5,<2.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'pyparsing>=2.4.0,<3',
 'python-dateutil>=2.8.1,<3.0.0',
 'pyyaml>=5.4.1,<6.0.0',
 'rich[jupyter]>=12.0.0,<13.0.0',
 'sqlmodel>=0.0.6,<0.1.0']

extras_require = \
{'server': ['fastapi>=0.75.0,<0.76.0', 'uvicorn[standard]>=0.17.5,<0.18.0'],
 'stacks': ['python-terraform>=0.10.1,<0.11.0']}

entry_points = \
{'console_scripts': ['zenml = zenml.cli.cli:cli']}

setup_kwargs = {
    'name': 'zenml',
    'version': '0.13.0',
    'description': 'ZenML: Write production-ready ML code.',
    'long_description': '<!-- PROJECT SHIELDS -->\n<!--\n*** I\'m using markdown "reference style" links for readability.\n*** Reference links are enclosed in brackets [ ] instead of parentheses ( ).\n*** See the bottom of this document for the declaration of the reference variables\n*** for contributors-url, forks-url, etc. This is an optional, concise syntax you may use.\n*** https://www.markdownguide.org/basic-syntax/#reference-style-links\n-->\n\n[![PyPi][pypi-shield]][pypi-url]\n[![PyPi][pypiversion-shield]][pypi-url]\n[![PyPi][downloads-shield]][downloads-url]\n[![Contributors][contributors-shield]][contributors-url]\n[![License][license-shield]][license-url]\n[![Build][build-shield]][build-url]\n[![Interrogate][interrogate-shield]][interrogate-url]\n<!-- [![CodeCov][codecov-shield]][codecov-url] -->\n\n<!-- MARKDOWN LINKS & IMAGES -->\n<!-- https://www.markdownguide.org/basic-syntax/#reference-style-links -->\n[pypi-shield]: https://img.shields.io/pypi/pyversions/zenml?style=for-the-badge\n[pypi-url]: https://pypi.org/project/zenml/\n[pypiversion-shield]: https://img.shields.io/pypi/v/zenml?style=for-the-badge\n\n[downloads-shield]: https://img.shields.io/pypi/dm/zenml?style=for-the-badge\n[downloads-url]: https://pypi.org/project/zenml/\n[codecov-shield]: https://img.shields.io/codecov/c/gh/zenml-io/zenml?style=for-the-badge\n[codecov-url]: https://codecov.io/gh/zenml-io/zenml\n[contributors-shield]: https://img.shields.io/github/contributors/zenml-io/zenml?style=for-the-badge\n[contributors-url]: https://github.com/othneildrew/Best-README-Template/graphs/contributors\n[license-shield]: https://img.shields.io/github/license/zenml-io/zenml?style=for-the-badge\n[license-url]: https://github.com/zenml-io/zenml/blob/main/LICENSE\n[linkedin-shield]: https://img.shields.io/badge/-LinkedIn-black.svg?style=for-the-badge&logo=linkedin&colorB=555\n[linkedin-url]: https://www.linkedin.com/company/zenml/\n[twitter-shield]: https://img.shields.io/twitter/follow/zenml_io?style=for-the-badge\n[twitter-url]: https://twitter.com/zenml_io\n[slack-shield]: https://img.shields.io/badge/-Slack-black.svg?style=for-the-badge&logo=linkedin&colorB=555\n[slack-url]: https://zenml.io/slack-invite\n[build-shield]: https://img.shields.io/github/workflow/status/zenml-io/zenml/Build,%20Lint,%20Unit%20&%20Integration%20Test/develop?logo=github&style=for-the-badge\n[build-url]: https://github.com/zenml-io/zenml/actions/workflows/ci.yml\n[interrogate-shield]: https://img.shields.io/badge/Interrogate-100%25-brightgreen?style=for-the-badge&logo=interrogate\n[interrogate-url]: https://interrogate.readthedocs.io/en/latest/\n\n\n<!-- PROJECT LOGO -->\n<br />\n<div align="center">\n  <a href="https://zenml.io">\n    <img src="docs/book/assets/zenml_logo.png" alt="Logo" width="400">\n  </a>\n\n  <h3 align="center">Build portable, production-ready MLOps pipelines.</h3>\n\n  <p align="center">\n    A simple yet powerful open-source framework that scales your MLOps stack with your needs.\n    <br />\n    <a href="https://docs.zenml.io/"><strong>Explore the docs »</strong></a>\n    <br />\n    <div align="center">\n      Join our <a href="https://zenml.io/slack-invite" target="_blank">\n      <img width="25" src="https://cdn3.iconfinder.com/data/icons/logos-and-brands-adobe/512/306_Slack-512.png" alt="Slack"/>\n    <b>Slack Community</b> </a> and be part of the ZenML family.\n    </div>\n    <br />\n    <a href="https://zenml.io/features">Features</a>\n    ·\n    <a href="https://zenml.io/roadmap">Roadmap</a>\n    ·\n    <a href="https://github.com/zenml-io/zenml/issues">Report Bug</a>\n    ·\n    <a href="https://zenml.io/discussion">Vote New Features</a>\n    ·\n    <a href="https://blog.zenml.io/">Read Blog</a>\n    ·\n    <a href="#-meet-the-team">Meet the Team</a>\n    <br />\n    🎉 Version 0.13.0 is out. Check out the release notes\n    <a href="https://github.com/zenml-io/zenml/releases">here</a>.\n    <br />\n    <br />\n    <a href="https://www.linkedin.com/company/zenml/">\n    <img src="https://img.shields.io/badge/-LinkedIn-black.svg?style=for-the-badge&logo=linkedin&colorB=555" alt="Logo">\n    </a>\n    <a href="https://twitter.com/zenml_io">\n    <img src="https://img.shields.io/badge/-Twitter-black.svg?style=for-the-badge&logo=twitter&colorB=555" alt="Logo">\n    </a>\n  </p>\n</div>\n\n<!-- TABLE OF CONTENTS -->\n<details>\n  <summary>🏁 Table of Contents</summary>\n  <ol>\n    <li>\n      <a href="#-why-zenml">Why ZenML?</a>\n    </li>\n    <li>\n    <a href="#-what-is-zenml">What is ZenML?</a>\n    </li>\n    <li>\n      <a href="#-getting-started">Getting Started</a>\n      <ul>\n        <li><a href="#-installation">Installation</a></li>\n        <li><a href="#-first-run">First run</a></li>\n        <li><a href="#-zenbytes">ZenBytes</a></li>\n        <li><a href="#-zenfiles">ZenFiles</a></li>\n      </ul>\n    </li>\n    <li><a href="#-collaborate-with-your-team">Collaborate with your team</a></li>\n    <li><a href="#-learn-more">Learn More</a></li>\n    <li><a href="#-roadmap">Roadmap</a></li>\n    <li><a href="#-contributing-and-community">Contributing and Community</a></li>\n    <li><a href="#-meet-the-team">Meet the Team</a></li>\n    <li><a href="#-getting-help">Getting Help</a></li>\n    <li><a href="#-license">License</a></li>\n  </ol>\n</details>\n\n<br />\n\n# 🤖 Why ZenML?\n\n🤹 Are you an ML engineer or data scientist shipping models to production and juggling a plethora of tools? \n\n🤷\u200d♂️ Do you struggle with versioning data, code, and models in your projects? \n\n👀 Have you had trouble replicating production pipelines and monitoring models in production?\n\n✅ If you answered yes to any of the above, ZenML is here to help with all that and more...\n\nEveryone loves to train ML models, but few talks about shipping them into production, and even fewer can do it well.\nAt ZenML, we believe the journey from model development to production doesn\'t need to be long and painful.\n\n![The long journey from experimentation to production.](docs/book/assets/1-pipeline-hard-reproduce.png)\n\n\nWith ZenML, you can concentrate on what you do best - developing ML models and not worry about infrastructure or deployment tools.\n\nIf you come from unstructured notebooks or scripts with lots of manual processes, ZenML will make the path to production easier and faster for you and your team.\nUsing ZenML allows you to own the entire pipeline - from experimentation to production.\n\nThis is why we built ZenML. Read more [here](https://blog.zenml.io/why-zenml/).\n\n\n\n\n# 💡 What is ZenML?\n\n<div align="center">\n    <img src="docs/book/assets/tailor.gif">\n</div>\n\n\nZenML is an extensible, open-source MLOps framework for creating portable, production-ready MLOps pipelines. It\'s built for Data Scientists, ML Engineers, and MLOps Developers to collaborate as they develop to production. \n\nZenML offers a simple and flexible syntax, is cloud- and tool-agnostic, and has interfaces/abstractions catered toward ML workflows. \nWith ZenML you\'ll have all your favorite tools in one place so you can tailor a workflow that caters to your specific needs.\n\n![ZenML unifies all your tools in one place.](docs/book/assets/sam-side-by-side-full-text.png)\n\nRead more on all tools you can readily use in the [integrations](https://zenml.io/integrations) section. Can\'t find your tool? You can always [write your own integration](https://docs.zenml.io/developer-guide/advanced-usage/custom-flavors) to use it with ZenML.\n\n# 🤸 Getting Started\n\n## 💾 Installation\n**Option 1** - Install ZenML via [PyPI](https://pypi.org/project/zenml/):\n\n```bash\npip install zenml\n```\n> **Note** - ZenML supports Python 3.7, 3.8, and 3.9.\n\n**Option 2** - If you’re feeling adventurous, try out the bleeding-edge installation:\n\n```bash\npip install git+https://github.com/zenml-io/zenml.git@develop --upgrade\n```\n\n> **Warning** - Fire dragons ahead. Proceed at your own risk!\n\n**Option 3** - Install via a Docker image hosted publicly on\n[DockerHub](https://hub.docker.com/r/zenmldocker/zenml):\n\n```shell\ndocker run -it zenmldocker/zenml /bin/bash\n```\n\n> **Warning** \n> #### Known installation issues for M1 Mac users\n>\n> If you have an M1 Mac machine and encounter an installation error, \n> try setting up `brew` and `pyenv` with Rosetta 2 and then install ZenML. The issue arises because some dependencies \n> aren’t fully compatible with the vanilla ARM64 Architecture. The following links may be helpful (Thank you @Reid Falconer) :\n>\n>- [Pyenv with Apple Silicon](http://sixty-north.com/blog/pyenv-apple-silicon.html)\n>- [Install Python Under Rosetta 2](https://medium.com/thinknum/how-to-install-python-under-rosetta-2-f98c0865e012)\n\n\n## 🏇 First run\n\nIf you\'re here for the first time, we recommend running:\n\n```shell\nzenml go\n```\n\nThis spins up a Jupyter notebook that walks you through various functionalities of ZenML at a high level.\n\nBy the end, you\'ll get a glimpse of how to use ZenML to:\n\n+ Train, evaluate, deploy, and embed a model in an inference pipeline.\n+ Automatically track and version data, models, and other artifacts.\n+ Track model hyperparameters and metrics with experiment tracking tools.\n+ Measure and visualize train-test skew, training-serving skew, and data drift.\n\n## 👨\u200d🍳 Open Source MLOps Stack Recipes\n\nZenML boasts a ton of [integrations](https://zenml.io/integrations) into popular MLOps tools. The [ZenML Stack](https://docs.zenml.io/developer-guide/stacks-profiles-repositories) concept ensures that these tools work nicely together, therefore bringing structure and standardization into the MLOps workflow.\n\nHowever, ZenML assumes that the stack infrastructure for these tools is already provisioned. If you do not have deployed infrastructure, and want to quickly spin up combinations of tools on the cloud, the [MLOps stack sister repository](https://github.com/zenml-io/mlops-stacks) contains a series of Terraform-based recipes to provision such stacks. These recipes can be used directly with ZenML:\n\n```bash\npip install zenml[stacks]\n\nzenml stack recipe deploy <NAME_OF_STACK_RECIPE> --import\n```\n\nThe above command not only provisions the given tools, but also automatically creates a ZenML stack with the configuration of the deployed recipe!\n\n## 🍰 ZenBytes\nNew to MLOps? Get up to speed by visiting the [ZenBytes](https://github.com/zenml-io/zenbytes) repo.\n\n>ZenBytes is a series of short practical MLOps lessons taught using ZenML. \n>It covers many of the [core concepts](https://docs.zenml.io/getting-started/core-concepts) widely used in ZenML and MLOps in general.\n\n## 📜 ZenFiles\nAlready comfortable with ZenML and wish to elevate your pipeline into production mode? Check out [ZenFiles](https://github.com/zenml-io/zenfiles).\n\n>ZenFiles is a collection of production-grade ML use-cases powered by ZenML. They are fully fleshed out, end-to-end projects that showcase ZenML\'s capabilities. They can also serve as a template from which to start similar projects.\n\n# 👭 Collaborate with your team\n\nZenML is built to support teams working together. \nThe underlying infrastructure on which your ML workflows run can be shared, as can the data, assets, and artifacts in your workflow. \n\nIn ZenML, a Stack represents a set of configurations for your MLOps tools and infrastructure. You can quickly share your ZenML stack with anyone by exporting the stack:\n\n```\nzenml stack export <STACK_NAME> <FILENAME.yaml>\n```\n\nSimilarly, you can import a stack by running:\n```\nzenml stack import <STACK_NAME> <FILENAME.yaml>\n```\n\nLearn more on importing/exporting stacks [here](https://docs.zenml.io/collaborate/stack-export-import).\n\n\nThe [ZenML Profiles](https://docs.zenml.io/collaborate/zenml-store) offer an easy way to manage and switch between your stacks. All your stacks, components, and other classes of ZenML objects can be stored in a central location and shared across multiple users, teams, and automated systems such as CI/CD processes.\n\nWith the [ZenServer](https://docs.zenml.io/collaborate/zenml-server) \nyou can deploy ZenML as a centralized service and connect entire teams and organizations to an easy-to-manage collaboration platform that provides a unified view of the MLOps processes, tools, and technologies that support your entire AI/ML project lifecycle.\n\nRead more about using ZenML for collaboration [here](https://docs.zenml.io/collaborate/collaborate-with-zenml).\n\n# 📖 Learn More\n\n| ZenML Resources | Description |\n| ------------- | - |\n| 🧘\u200d♀️ **[ZenML 101]** | New to ZenML? Here\'s everything you need to know! |\n| ⚛️ **[Core Concepts]** | Some key terms and concepts we use. |\n| 🚀 **[Our latest release]** | New features, bug fixes. |\n| 🗳 **[Vote for Features]** | Pick what we work on next! |\n| 📓 **[Docs]** | Full documentation for creating your own ZenML pipelines. |\n| 📒 **[API Reference]** | Detailed reference on ZenML\'s API. |\n| 🍰 **[ZenBytes]** | A guided and in-depth tutorial on MLOps and ZenML. |\n| 🗂️️ **[ZenFiles]** | End-to-end projects using ZenML. |\n| 👨\u200d🍳 **[MLOps Stacks]** | Terraform based infrastructure recipes for pre-made ZenML stacks. |\n| ⚽️ **[Examples]** | Learn best through examples where ZenML is used? We\'ve got you covered. |\n| 📬 **[Blog]** | Use cases of ZenML and technical deep dives on how we built it. |\n| 🔈 **[Podcast]** | Conversations with leaders in ML, released every 2 weeks. |\n| 📣 **[Newsletter]** | We build ZenML in public. Subscribe to learn how we work. |\n| 💬 **[Join Slack]** | Need help with your specific use case? Say hi on Slack! |\n| 🗺 **[Roadmap]** | See where ZenML is working to build new features. |\n| 🙋\u200d♀️ **[Contribute]** | How to contribute to the ZenML project and code base. |\n\n[ZenML 101]: https://docs.zenml.io/\n[Core Concepts]: https://docs.zenml.io/getting-started/core-concepts\n[API Guide]: https://docs.zenml.io/v/docs/developer-guide/steps-and-pipelines/functional-vs-class-based-api\n[Our latest release]: https://github.com/zenml-io/zenml/releases\n[Vote for Features]: https://zenml.io/discussion\n[Docs]: https://docs.zenml.io/\n[API Reference]: https://apidocs.zenml.io/\n[ZenBytes]: https://github.com/zenml-io/zenbytes\n[ZenFiles]: https://github.com/zenml-io/zenfiles\n[MLOps Stacks]: https://github.com/zenml-io/mlops-stacks\n[Examples]: https://github.com/zenml-io/zenml/tree/main/examples\n[Blog]: https://blog.zenml.io/\n[Podcast]: https://podcast.zenml.io/\n[Newsletter]: https://zenml.io/newsletter/\n[Join Slack]: https://zenml.io/slack-invite/\n[Roadmap]: https://zenml.io/roadmap\n[Contribute]: https://github.com/zenml-io/zenml/blob/main/CONTRIBUTING.md\n\n\n# 🗺 Roadmap\n\nZenML is being built in public. The [roadmap](https://zenml.io/roadmap) is a\nregularly updated source of truth for the ZenML community to understand where\nthe product is going in the short, medium, and long term.\n\nZenML is managed by a [core team](https://zenml.io/company#CompanyTeam) of developers that are\nresponsible for making key decisions and incorporating feedback from the\ncommunity. The team oversees feedback via various channels, and you can directly\ninfluence the roadmap as follows:\n\n- Vote on your most wanted feature on our [Discussion\n  board](https://zenml.io/discussion).\n- Start a thread in our [Slack channel](https://zenml.io/slack-invite).\n- [Create an issue](https://github.com/zenml-io/zenml/issues/new/choose) on our Github repo.\n\n\n\n# 🙌 Contributing and Community\n\nWe would love to develop ZenML together with our community! Best way to get\nstarted is to select any issue from the [`good-first-issue`\nlabel](https://github.com/zenml-io/zenml/labels/good%20first%20issue). If you\nwould like to contribute, please review our [Contributing\nGuide](CONTRIBUTING.md) for all relevant details.\n\n<br>\n\n![Repobeats analytics\nimage](https://repobeats.axiom.co/api/embed/635c57b743efe649cadceba6a2e6a956663f96dd.svg\n"Repobeats analytics image")\n\n\n# 👩\u200d👩\u200d👧\u200d👦 Meet the Team\n\n![Meet the Team](./docs/book/assets/meet_the_team.jpeg)\n\nHave a question that\'s too hard to express on our Slack? Is it just too much effort to say everything on a \nlong GitHub issue? Or are you just curious about what ZenML has been up to in the past week? Well, register now for the ZenML Office (Half) Hour to get your answers and more!\nIt\'s free and open to everyone.\n\nEvery week, part of the ZenML [core team](https://zenml.io/company#CompanyTeam) will pop in for 30 minutes to interact directly with the community. Sometimes we\'ll be presenting a feature. Other times we just take questions and have fun. Join us if you are curious about ZenML, or just want to talk shop about MLOps.\n\n\n\nWe will host the gathering every Wednesday 8:30AM PT (5:30PM CET). \nRegister now through [this link](https://www.eventbrite.com/e/zenml-meet-the-community-tickets-354426688767), \nor subscribe to the [public events calendar](https://calendar.google.com/calendar/u/0/r?cid=Y19iaDJ0Zm44ZzdodXBlbnBzaWplY3UwMmNjZ0Bncm91cC5jYWxlbmRhci5nb29nbGUuY29t) to get notified \nbefore every community gathering.\n\n# 🆘 Getting Help\n\nThe first point of call should be [our Slack group](https://zenml.io/slack-invite/).\nAsk your questions about bugs or specific use cases, and someone from the [core team](https://zenml.io/company#CompanyTeam) will respond.\nOr, if you prefer, [open an issue](https://github.com/zenml-io/zenml/issues/new/choose) on our GitHub repo.\n\n\n# 📜 License\n\nZenML is distributed under the terms of the Apache License Version 2.0. \nA complete version of the license is available in the [LICENSE](LICENSE) file in\nthis repository. Any contribution made to this project will be licensed under\nthe Apache License Version 2.0.\n',
    'author': 'ZenML GmbH',
    'author_email': 'info@zenml.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://zenml.io',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<3.10',
}


setup(**setup_kwargs)



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Entrypoint configuration for the Kubernetes master/orchestrator pod."""

import json
from typing import Dict, List, Set, Tuple

from tfx.proto.orchestration.pipeline_pb2 import Pipeline as Pb2Pipeline

from zenml.entrypoints.step_entrypoint_configuration import (
    INPUT_ARTIFACT_SOURCES_OPTION,
    MATERIALIZER_SOURCES_OPTION,
    STEP_SOURCE_OPTION,
)
from zenml.integrations.kubernetes.orchestrators.kubernetes_step_entrypoint_configuration import (
    RUN_NAME_OPTION,
    KubernetesStepEntrypointConfiguration,
)
from zenml.steps.base_step import BaseStep

PIPELINE_NAME_OPTION = "pipeline_name"
IMAGE_NAME_OPTION = "image_name"
NAMESPACE_OPTION = "kubernetes_namespace"
PIPELINE_CONFIG_OPTION = "pipeline_config"

STEP_SPECIFIC_STEP_ENTRYPOINT_OPTIONS = [
    STEP_SOURCE_OPTION,
    INPUT_ARTIFACT_SOURCES_OPTION,
    MATERIALIZER_SOURCES_OPTION,
]  # options from StepEntrypointConfiguration that change from step to step.


def split_step_args(step_args: List[str]) -> Tuple[List[str], List[str]]:
    """Split step args into fixed and step-specific.

    We want to have them separate so we can send the fixed args to the
    orchestrator pod only once.

    Args:
        step_args: list of ALL step args.
            E.g. ["--arg1", "arg1_value", "--arg2", "arg2_value", ...].

    Returns:
        Tuple (fixed step args, step-specific args).
    """
    fixed_args = []
    step_specific_args = []
    for i, arg in enumerate(step_args):
        if not arg.startswith("--"):  # arg is a value, not an option
            continue
        option_and_value = step_args[i : i + 2]  # e.g. ["--name", "Aria"]
        is_fixed = arg[2:] not in STEP_SPECIFIC_STEP_ENTRYPOINT_OPTIONS
        if is_fixed:
            fixed_args += option_and_value
        else:
            step_specific_args += option_and_value
    return fixed_args, step_specific_args


class KubernetesOrchestratorEntrypointConfiguration:
    """Entrypoint configuration for the k8s master/orchestrator pod."""

    @classmethod
    def get_entrypoint_options(cls) -> Set[str]:
        """Gets all the options required for running this entrypoint.

        Returns:
            Entrypoint options.
        """
        options = {
            RUN_NAME_OPTION,
            PIPELINE_NAME_OPTION,
            IMAGE_NAME_OPTION,
            NAMESPACE_OPTION,
            PIPELINE_CONFIG_OPTION,
        }
        return options

    @classmethod
    def get_entrypoint_command(cls) -> List[str]:
        """Returns a command that runs the entrypoint module.

        Returns:
            Entrypoint command.
        """
        command = [
            "python",
            "-m",
            "zenml.integrations.kubernetes.orchestrators.kubernetes_orchestrator_entrypoint",
        ]
        return command

    @classmethod
    def get_entrypoint_arguments(
        cls,
        run_name: str,
        pipeline_name: str,
        image_name: str,
        kubernetes_namespace: str,
        pb2_pipeline: Pb2Pipeline,
        sorted_steps: List[BaseStep],
        pipeline_dag: Dict[str, List[str]],
    ) -> List[str]:
        """Gets all arguments that the entrypoint command should be called with.

        Args:
            run_name: Name of the ZenML run.
            pipeline_name: Name of the ZenML pipeline.
            image_name: Name of the Docker image.
            kubernetes_namespace: Name of the Kubernetes namespace.
            pb2_pipeline: ZenML pipeline in TFX pb2 format.
            sorted_steps: List of steps in execution order.
            pipeline_dag: For each step, list of steps that need to run before.

        Returns:
            List of entrypoint arguments.
        """

        def _get_step_args(step: BaseStep) -> List[str]:
            """Get the entrypoint args for a specific step.

            Args:
                step: ZenML step for which to get entrypoint args.

            Returns:
                Entrypoint args of the step.
            """
            return (
                KubernetesStepEntrypointConfiguration.get_entrypoint_arguments(
                    step=step,
                    pb2_pipeline=pb2_pipeline,
                    **{RUN_NAME_OPTION: run_name},
                )
            )

        # Get name, command, and args for each step
        step_names = [step.name for step in sorted_steps]
        step_command = (
            KubernetesStepEntrypointConfiguration.get_entrypoint_command()
        )
        fixed_step_args = []
        if len(sorted_steps) > 0:
            first_step_args = _get_step_args(sorted_steps[0])
            fixed_step_args = split_step_args(first_step_args)[0]
        step_specific_args = {
            step.name: split_step_args(_get_step_args(step))[1]
            for step in sorted_steps
        }  # e.g.: {"trainer": train_step_args, ...}

        # Serialize all complex datatype args into a single JSON string
        pipeline_config = {
            "sorted_steps": step_names,
            "step_command": step_command,
            "fixed_step_args": fixed_step_args,
            "step_specific_args": step_specific_args,
            "pipeline_dag": pipeline_dag,
        }
        pipeline_config_json = json.dumps(pipeline_config)

        # Define entrypoint args.
        args = [
            f"--{RUN_NAME_OPTION}",
            run_name,
            f"--{PIPELINE_NAME_OPTION}",
            pipeline_name,
            f"--{IMAGE_NAME_OPTION}",
            image_name,
            f"--{NAMESPACE_OPTION}",
            kubernetes_namespace,
            f"--{PIPELINE_CONFIG_OPTION}",
            pipeline_config_json,
        ]

        return args



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the Azure Secrets Manager integration."""

import base64
import json
import re
from typing import Any, ClassVar, List, Optional

from azure.core.exceptions import ResourceNotFoundError
from azure.identity import DefaultAzureCredential
from azure.keyvault.secrets import SecretClient

from zenml.exceptions import SecretExistsError
from zenml.integrations.azure import AZURE_SECRETS_MANAGER_FLAVOR
from zenml.logger import get_logger
from zenml.secret.base_secret import BaseSecretSchema
from zenml.secret.secret_schema_class_registry import SecretSchemaClassRegistry
from zenml.secrets_managers.base_secrets_manager import (
    ZENML_SECRET_NAME_LABEL,
    BaseSecretsManager,
    SecretsManagerScope,
)
from zenml.secrets_managers.utils import secret_from_dict, secret_to_dict

logger = get_logger(__name__)

ZENML_SCHEMA_NAME = "zenml-schema-name"
ZENML_GROUP_KEY = "zenml-group-key"
ZENML_KEY_NAME = "zenml-key-name"
ZENML_AZURE_SECRET_SCOPE_PATH_SEPARATOR = "-"


class AzureSecretsManager(BaseSecretsManager):
    """Class to interact with the Azure secrets manager.

    Attributes:
        key_vault_name: Name of an Azure Key Vault that this secrets manager
            will use to store secrets.
    """

    key_vault_name: str

    # Class configuration
    FLAVOR: ClassVar[str] = AZURE_SECRETS_MANAGER_FLAVOR
    SUPPORTS_SCOPING: ClassVar[bool] = True
    CLIENT: ClassVar[Any] = None

    @classmethod
    def _ensure_client_connected(cls, vault_name: str) -> None:
        if cls.CLIENT is None:
            KVUri = f"https://{vault_name}.vault.azure.net"

            credential = DefaultAzureCredential()
            cls.CLIENT = SecretClient(vault_url=KVUri, credential=credential)

    @classmethod
    def _validate_scope(
        cls,
        scope: SecretsManagerScope,
        namespace: Optional[str],
    ) -> None:
        """Validate the scope and namespace value.

        Args:
            scope: Scope value.
            namespace: Optional namespace value.
        """
        if namespace:
            cls.validate_secret_name_or_namespace(namespace, scope)

    @classmethod
    def validate_secret_name_or_namespace(
        cls,
        name: str,
        scope: SecretsManagerScope,
    ) -> None:
        """Validate a secret name or namespace.

        Azure secret names must contain only alphanumeric characters and the
        character `-`.

        Given that we also save secret names and namespaces as labels, we are
        also limited by the 256 maximum size limitation that Azure imposes on
        label values. An arbitrary length of 100 characters is used here for
        the maximum size for the secret name and namespace.

        Args:
            name: the secret name or namespace
            scope: the current scope

        Raises:
            ValueError: if the secret name or namespace is invalid
        """
        if scope == SecretsManagerScope.NONE:
            # to preserve backwards compatibility, we don't validate the
            # secret name for unscoped secrets.
            return

        if not re.fullmatch(r"[0-9a-zA-Z-]+", name):
            raise ValueError(
                f"Invalid secret name or namespace '{name}'. Must contain "
                f"only alphanumeric characters and the character -."
            )

        if len(name) > 100:
            raise ValueError(
                f"Invalid secret name or namespace '{name}'. The length is "
                f"limited to maximum 100 characters."
            )

    def validate_secret_name(self, name: str) -> None:
        """Validate a secret name.

        Args:
            name: the secret name
        """
        self.validate_secret_name_or_namespace(name, self.scope)

    def _create_or_update_secret(self, secret: BaseSecretSchema) -> None:
        """Creates a new secret or updated an existing one.

        Args:
            secret: the secret to register or update
        """
        if self.scope == SecretsManagerScope.NONE:
            # legacy, non-scoped secrets

            for key, value in secret.content.items():
                encoded_key = base64.b64encode(
                    f"{secret.name}-{key}".encode()
                ).hex()
                azure_secret_name = f"zenml-{encoded_key}"

                self.CLIENT.set_secret(azure_secret_name, value)
                self.CLIENT.update_secret_properties(
                    azure_secret_name,
                    tags={
                        ZENML_GROUP_KEY: secret.name,
                        ZENML_KEY_NAME: key,
                        ZENML_SCHEMA_NAME: secret.TYPE,
                    },
                )

                logger.debug(
                    "Secret `%s` written to the Azure Key Vault.",
                    azure_secret_name,
                )
        else:
            azure_secret_name = self._get_scoped_secret_name(
                secret.name,
                separator=ZENML_AZURE_SECRET_SCOPE_PATH_SEPARATOR,
            )
            self.CLIENT.set_secret(
                azure_secret_name,
                json.dumps(secret_to_dict(secret)),
            )
            self.CLIENT.update_secret_properties(
                azure_secret_name,
                tags=self._get_secret_metadata(secret),
            )

    def register_secret(self, secret: BaseSecretSchema) -> None:
        """Registers a new secret.

        Args:
            secret: the secret to register

        Raises:
            SecretExistsError: if the secret already exists
        """
        self.validate_secret_name(secret.name)
        self._ensure_client_connected(self.key_vault_name)

        if secret.name in self.get_all_secret_keys():
            raise SecretExistsError(
                f"A Secret with the name '{secret.name}' already exists."
            )

        self._create_or_update_secret(secret)

    def get_secret(self, secret_name: str) -> BaseSecretSchema:
        """Get a secret by its name.

        Args:
            secret_name: the name of the secret to get

        Returns:
            The secret.

        Raises:
            KeyError: if the secret does not exist
            ValueError: if the secret is named 'name'
        """
        self.validate_secret_name(secret_name)
        self._ensure_client_connected(self.key_vault_name)
        zenml_secret: Optional[BaseSecretSchema] = None

        if self.scope == SecretsManagerScope.NONE:
            # Legacy secrets are mapped to multiple Azure secrets, one for
            # each secret key

            secret_contents = {}
            zenml_schema_name = ""

            for secret_property in self.CLIENT.list_properties_of_secrets():
                tags = secret_property.tags

                if tags and tags.get(ZENML_GROUP_KEY) == secret_name:
                    secret_key = tags.get(ZENML_KEY_NAME)
                    if not secret_key:
                        raise ValueError("Missing secret key tag.")

                    if secret_key == "name":
                        raise ValueError("The secret's key cannot be 'name'.")

                    response = self.CLIENT.get_secret(secret_property.name)
                    secret_contents[secret_key] = response.value

                    zenml_schema_name = tags.get(ZENML_SCHEMA_NAME)

            if secret_contents:
                secret_contents["name"] = secret_name

                secret_schema = SecretSchemaClassRegistry.get_class(
                    secret_schema=zenml_schema_name
                )
                zenml_secret = secret_schema(**secret_contents)
        else:
            # Scoped secrets are mapped 1-to-1 with Azure secrets

            try:
                response = self.CLIENT.get_secret(
                    self._get_scoped_secret_name(
                        secret_name,
                        separator=ZENML_AZURE_SECRET_SCOPE_PATH_SEPARATOR,
                    ),
                )

                scope_tags = self._get_secret_scope_metadata(secret_name)

                # all scope tags need to be included in the Azure secret tags,
                # otherwise the secret does not belong to the current scope,
                # even if it has the same name
                if scope_tags.items() <= response.properties.tags.items():
                    zenml_secret = secret_from_dict(
                        json.loads(response.value), secret_name=secret_name
                    )
            except ResourceNotFoundError:
                pass

        if not zenml_secret:
            raise KeyError(f"Can't find the specified secret '{secret_name}'")

        return zenml_secret

    def get_all_secret_keys(self) -> List[str]:
        """Get all secret keys.

        Returns:
            A list of all secret keys
        """
        self._ensure_client_connected(self.key_vault_name)

        set_of_secrets = set()

        for secret_property in self.CLIENT.list_properties_of_secrets():
            tags = secret_property.tags
            if not tags:
                continue

            if self.scope == SecretsManagerScope.NONE:
                # legacy, non-scoped secrets
                if ZENML_GROUP_KEY in tags:
                    set_of_secrets.add(tags.get(ZENML_GROUP_KEY))
                continue

            scope_tags = self._get_secret_scope_metadata()
            # all scope tags need to be included in the Azure secret tags,
            # otherwise the secret does not belong to the current scope
            if scope_tags.items() <= tags.items():
                set_of_secrets.add(tags.get(ZENML_SECRET_NAME_LABEL))

        return list(set_of_secrets)

    def update_secret(self, secret: BaseSecretSchema) -> None:
        """Update an existing secret by creating new versions of the existing secrets.

        Args:
            secret: the secret to update

        Raises:
            KeyError: if the secret does not exist
        """
        self.validate_secret_name(secret.name)
        self._ensure_client_connected(self.key_vault_name)

        if secret.name not in self.get_all_secret_keys():
            raise KeyError(f"Can't find the specified secret '{secret.name}'")

        self._create_or_update_secret(secret)

    def delete_secret(self, secret_name: str) -> None:
        """Delete an existing secret. by name.

        Args:
            secret_name: the name of the secret to delete

        Raises:
            KeyError: if the secret no longer exists
        """
        self.validate_secret_name(secret_name)
        self._ensure_client_connected(self.key_vault_name)

        if self.scope == SecretsManagerScope.NONE:
            # legacy, non-scoped secrets

            # Go through all Azure secrets and delete the ones with the
            # secret_name as label.
            for secret_property in self.CLIENT.list_properties_of_secrets():
                tags = secret_property.tags
                if tags and tags.get(ZENML_GROUP_KEY) == secret_name:
                    self.CLIENT.begin_delete_secret(
                        secret_property.name
                    ).result()

        else:
            if secret_name not in self.get_all_secret_keys():
                raise KeyError(
                    f"Can't find the specified secret '{secret_name}'"
                )
            self.CLIENT.begin_delete_secret(
                self._get_scoped_secret_name(
                    secret_name,
                    separator=ZENML_AZURE_SECRET_SCOPE_PATH_SEPARATOR,
                ),
            ).result()

    def delete_all_secrets(self) -> None:
        """Delete all existing secrets."""
        self._ensure_client_connected(self.key_vault_name)

        # List all secrets.
        for secret_property in self.CLIENT.list_properties_of_secrets():

            tags = secret_property.tags
            if not tags:
                continue

            if self.scope == SecretsManagerScope.NONE:
                # legacy, non-scoped secrets
                if ZENML_GROUP_KEY in tags:
                    logger.info(
                        "Deleted key-value pair {`%s`, `***`} from secret "
                        "`%s`",
                        secret_property.name,
                        tags.get(ZENML_GROUP_KEY),
                    )
                    self.CLIENT.begin_delete_secret(
                        secret_property.name
                    ).result()
                continue

            scope_tags = self._get_secret_scope_metadata()
            # all scope tags need to be included in the Azure secret tags,
            # otherwise the secret does not belong to the current scope
            if scope_tags.items() <= tags.items():
                self.CLIENT.begin_delete_secret(secret_property.name).result()



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the GitHub Container Registry."""

from typing import ClassVar

from zenml.container_registries.base_container_registry import (
    BaseContainerRegistry,
)
from zenml.enums import ContainerRegistryFlavor


class GitHubContainerRegistry(BaseContainerRegistry):
    """Class for GitHub Container Registry.

    Attributes:
        automatic_token_authentication: If `True`, use automatic token
            authentication (https://docs.github.com/en/actions/security-guides/automatic-token-authentication#using-the-github_token-in-a-workflow)
            when trying to access this container registry from within a GitHub
            Actions environment.
    """

    automatic_token_authentication: bool = False

    # Class Configuration
    FLAVOR: ClassVar[str] = ContainerRegistryFlavor.GITHUB



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Stack wrapper implementation."""

from typing import List, Optional
from uuid import UUID

from pydantic import BaseModel

from zenml.enums import StackComponentType
from zenml.stack import Stack
from zenml.zen_stores.models import ComponentWrapper


class StackWrapper(BaseModel):
    """Network Serializable Wrapper describing a Stack."""

    name: str
    components: List[ComponentWrapper]

    @classmethod
    def from_stack(cls, stack: Stack) -> "StackWrapper":
        """Creates a StackWrapper from an actual Stack instance.

        Args:
            stack: the instance of a Stack

        Returns:
            a StackWrapper
        """
        return cls(
            name=stack.name,
            components=[
                ComponentWrapper.from_component(component)
                for t, component in stack.components.items()
            ],
        )

    def to_stack(self) -> Stack:
        """Creates the corresponding Stack instance from the wrapper.

        Returns:
            the corresponding Stack instance
        """
        stack_components = {}
        for component_wrapper in self.components:
            component_type = component_wrapper.type
            component = component_wrapper.to_component()
            stack_components[component_type] = component

        return Stack.from_components(
            name=self.name, components=stack_components
        )

    def get_component_wrapper(
        self, component_type: StackComponentType
    ) -> Optional[ComponentWrapper]:
        """Returns the component of the given type.

        Args:
            component_type: the type of the component to return

        Returns:
            the component of the given type or None if not found
        """
        for component_wrapper in self.components:
            if component_wrapper.type == component_type:
                return component_wrapper

        return None


class StoreAssociation(BaseModel):
    """Model for the association between an artifact store and a metadata store.

    Args:
        artifact_store_uuid: The UUID of the artifact store.
        metadata_store_uuid: The UUID of the metadata store.
    """

    artifact_store_uuid: UUID
    metadata_store_uuid: UUID
